"""Provide the DeletableMixin class."""

from . import BaseModel


class DeletableMixin(BaseModel):
    """Interface for classes that can be deleted."""

    def delete(self):
        """Delete the object.

        Example usage:

        .. code-block:: python

            bot = credential_manager.bot("name")

            reddit_app = bot.reddit_app
            reddit_app.delete()

            sentry_token = bot.sentry_token
            sentry_token.delete()

            database_credential = bot.database_credential
            database_credential.delete()

            user_verification = cre

        """
        self._credmgr.delete(f"{self._path}/{self.id}")
        del self



import pytest

from credmgr.exceptions import Conflict, NotFound, ServerError
from credmgr.models import SentryToken


def test_create_sentry_token(recorder, credential_manager):
    data = {"name": "test_sentry_token", "dsn": "https://key@sentry.jesassn.org/id"}
    sentry_token = credential_manager.sentry_token.create(**data)
    for key, value in data.items():
        assert getattr(sentry_token, key) == value


def test_create_sentry_token_other_user(recorder, credential_manager):
    data = {
        "name": "test_sentry_token",
        "dsn": "https://key@sentry.jesassn.org/id",
        "owner": 4,
    }
    sentry_token = credential_manager.sentry_token.create(**data)
    for key, value in data.items():
        if key == "owner":
            assert getattr(sentry_token, key).id == value
            continue
        assert getattr(sentry_token, key) == value


def test_create_sentry_token_bad_params(recorder, credential_manager):
    data = {"name": "se", "dsn": "https://key@sentry.jesassn.org/id2"}
    with pytest.raises(ServerError):
        _ = credential_manager.sentry_token.create(**data)


def test_create_sentry_token_existing(recorder, credential_manager):
    data = {"name": "sentry_token", "dsn": "https://key@sentry.jesassn.org/id"}
    with pytest.raises(Conflict):
        _ = credential_manager.sentry_token.create(**data)


def test_delete_sentry_token(recorder, credential_manager):
    sentry_token = credential_manager.sentry_token(1)
    sentry_token.delete()
    with pytest.raises(NotFound):
        _ = credential_manager.sentry_token(1)


def test_edit_sentry_token(recorder, credential_manager):
    sentry_token = credential_manager.sentry_token(2)
    sentry_token.edit(dsn="https://key@sentry.jesassn.org/id_new")
    assert sentry_token.dsn == "https://key@sentry.jesassn.org/id_new"


def test_edit_sentry_token_conflicting_data(recorder, credential_manager):
    sentry_token = credential_manager.sentry_token(2)
    with pytest.raises(Conflict):
        sentry_token.edit(name="sentry_token")


def test_list_sentry_tokens(recorder, credential_manager):
    sentry_tokens = credential_manager.sentry_tokens()
    for sentry_token in sentry_tokens:
        assert isinstance(sentry_token, SentryToken)



"""Provide the UserVerification class."""

import json

from ..deprecation_decorator import warn_camel_case
from ..mixins import BaseApp, RedditAppMixin
from . import RedditApp
from .utils import _resolve_model_from_input, _resolve_user


class UserVerification(BaseApp, RedditAppMixin):
    """A class for UserVerification instances."""

    _can_fetch_by_name = True
    _credmgr_callable = "user_verification"
    _editable_attrs = ["user_id", "redditor", "reddit_app_id", "extra_data"]
    _get_by_name_path = "get_redditor"
    _name_attr = "user_id"
    _path = "/user_verifications"

    @classmethod
    def schema(cls):
        """Return the schema for this model."""
        if cls._schema is None:
            from .schemas import UserVerificationSchema

            cls._schema = UserVerificationSchema()
        return cls._schema

    def __init__(self, credmgr, **kwargs):
        """Initialize an User Verification instance.

        User verifications are for associating an unique ID with a redditor.

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param id: ID of this user verification.
        :param user_id: An unique ID to associate with a redditor.
        :param redditor: Redditor username to associate with an unique ID.
        :param reddit_app_id: ID of the :class:`.RedditApp` associated with this user
            verification.
        :param extra_data: Extra data to be associated with this user verification.
        :param owner_id: ID of the :class:`.User` that owns this user verification.

        """
        super().__init__(credmgr, **kwargs)

    @staticmethod
    @_resolve_user
    @warn_camel_case("userId", "redditApp", "extraData")
    def _create(
        _credmgr, user_id, reddit_app, redditor=None, extra_data=None, owner=None
    ):
        """Create a new User Verification.

        User verifications are for associating an unique ID with a redditor.

        :param str user_id: An unique ID to associate with a redditor.
        :param Union[RedditApp,int,str] reddit_app: Reddit app the user verification is
            for
        :param str redditor: Redditor the user verification is for. This is not usually
            set manually.
        :param dict extra_data: Extra JSON data to include with verification
        :param Union[User,int,str] owner: Owner of the verification. Requires admin to
            create for other users.

        :returns: The new :class:`.UserVerification` instance.

        """
        data = {
            "user_id": user_id,
            "reddit_app_id": _resolve_model_from_input(_credmgr, RedditApp, reddit_app),
        }
        if redditor:
            data["redditor"] = redditor
        if extra_data:
            data["extra_data"] = json.dumps(extra_data)
        if owner:
            data["owner_id"] = owner
        return _credmgr.post("/user_verifications", data=data)



import os


def capital_camel_case(path):
    return "".join(
        [i.capitalize() for i in os.path.basename(path).split(".")[0].split("_")]
    )


def gen_cassette_name():
    current_test = os.environ.get("PYTEST_CURRENT_TEST")
    test_dir, test_name = current_test.split("::")
    test_dir = capital_camel_case(os.path.basename(test_dir))
    test_name = test_name.split(" ")[0]
    return f"{test_dir}.{test_name}"



import pytest
from betamax import Betamax


def test_base_equal(recorder, credential_manager):
    bot = credential_manager.bot(1)
    bot2 = credential_manager.bot(1)
    assert bot == bot2


def test_base_not_equal(recorder, credential_manager):
    bot = credential_manager.bot(1)
    bot2 = credential_manager.bot(4)
    assert bot != bot2


def test_base_equal_different_type(recorder, credential_manager):
    item = credential_manager.bot(1)
    item2 = item.reddit_app
    assert item != item2


def test_base_camel_case(credential_manager):
    with Betamax(credential_manager._requestor._session).use_cassette(
        "TestBase.test_base_equal_different_type"
    ):
        item = credential_manager.bot(1)
        reddit_app = item.reddit_app
        with pytest.deprecated_call():
            redditApp = item.redditApp
        assert reddit_app == redditApp



"""Provide the EditableMixin class."""

from ..deprecation_decorator import warn_camel_case
from . import BaseModel


class EditableMixin(BaseModel):
    """Interface for classes that are editable."""

    _editable_attrs = []

    @warn_camel_case()
    def edit(self, **kwargs):
        """Edit the object.

        :param kwargs: The params to update on the object.

        :returns: The edited object.

        """
        payload = []

        for attr in self._editable_attrs:
            if attr in kwargs:
                path = f"/{self._serialize_mapping.get(attr, attr)}"
                payload.append({"op": "replace", "path": path, "value": kwargs[attr]})
        updated_object = self._credmgr.patch(f"{self._path}/{self.id}", data=payload)
        from . import BaseApp

        for key, value in updated_object.__dict__.items():
            if key in self._editable_attrs or (
                issubclass(type(value), BaseApp) and f"{key}_id" in self._editable_attrs
            ):
                setattr(self, key, value)
        return self



import json
import os
import socket
from sys import platform

import betamax
import pytest
from betamax import Betamax
from betamax_serializers import pretty_json

from credmgr import CredentialManager
from tests.utils import gen_cassette_name


def filter_refresh_token(interaction, current_cassette):  # pragma: no cover
    filter_items = ["refresh_token"]
    response = interaction.data["response"]
    body = response["body"]["string"]
    for item in filter_items:
        try:
            value = json.loads(body)[item]
            current_cassette.placeholders.append(
                betamax.cassette.cassette.Placeholder(
                    placeholder=f"<{item.upper()}>", replace=value
                )
            )
        except (KeyError, TypeError, ValueError):
            continue


def env_default(key):
    return os.environ.get(f"test_{key}", f"placeholder_{key}")


placeholders = {x: env_default(x) for x in ["api_token", "password", "username"]}

betamax.Betamax.register_serializer(pretty_json.PrettyJSONSerializer)
with betamax.Betamax.configure() as config:
    config.cassette_library_dir = f"{os.path.dirname(__file__)}/cassettes"
    config.default_cassette_options["serialize_with"] = "prettyjson"
    config.before_record(callback=filter_refresh_token)
    for key, value in placeholders.items():
        config.define_cassette_placeholder(f"<{key.upper()}>", value)


class Placeholders:
    def __init__(self, _dict):
        self.__dict__ = _dict


def pytest_configure():
    pytest.placeholders = Placeholders(placeholders)


if platform == "darwin":
    socket.gethostbyname = lambda x: "127.0.0.1"


@pytest.fixture()
def credential_manager():
    yield CredentialManager(api_token=pytest.placeholders.api_token)


@pytest.fixture()
def recorder(credential_manager):
    with Betamax(credential_manager._requestor._session).use_cassette(
        gen_cassette_name()
    ):
        yield



"""Provide the RedditApp class."""

import base64
from typing import TYPE_CHECKING, Union

import praw

from ..deprecation_decorator import warn_camel_case
from ..mixins import BaseApp
from .utils import _resolve_model_from_input, _resolve_user

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class RedditApp(BaseApp):
    """A class for Reddit Apps.

    To obtain an instance of this class execute:

    .. code-block:: python

        bot = credmgr.bot("BotName")
        reddit_app = bot.reddit_app

    """

    _credmgr_callable = "reddit_app"
    _editable_attrs = BaseApp._editable_attrs + [
        "app_description",
        "app_type",
        "client_id",
        "client_secret",
        "redirect_uri",
        "user_agent",
    ]
    _path = "/reddit_apps"
    _reddit = None

    @classmethod
    def schema(cls):
        """Return the schema for this model."""
        if cls._schema is None:
            from .schemas import RedditAppSchema

            cls._schema = RedditAppSchema()
        return cls._schema

    def __init__(self, credmgr, **kwargs):
        """Initialize a Reddit App instance.

        Reddit Apps are used for interacting with reddit

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param int id: ID of this Reddit App.
        :param str name: Name of this Reddit App.
        :param str client_id: Client ID of this Reddit App.
        :param str client_secret: Client secret of the Reddit App.
        :param str app_description: Description of what this Reddit App is used for.
        :param str user_agent: The user agent used to identify this Reddit App to
            Reddit.
        :param str app_type: Type of app this Reddit App is. One of: ``web``,
            ``installed``, or ``script``.
        :param str redirect_uri: URL that redditors are redirected to after authorizing
            this Reddit App to access their account.
        :param str state: Used to identify this Reddit App during the OAuth2 flow.
        :param int owner_id: ID of the `.User` that owns this Reddit App.

        """
        super().__init__(credmgr, **kwargs)

        if "state" in kwargs:
            self._fetched = True

    @staticmethod
    @_resolve_user
    @warn_camel_case(
        "clientId",
        "userAgent",
        "appType",
        "redirectUri",
        "clientSecret",
        "appDescription",
    )
    def _create(
        _credmgr,
        name,
        client_id,
        user_agent,
        app_type,
        redirect_uri,
        client_secret,
        app_description,
        enabled,
        owner=None,
    ):
        """Create a new Reddit App.

        Reddit Apps are used for interacting with Reddit

        :param str name: (required)
        :param str client_id: Client ID of the Reddit App (required)
        :param str user_agent: User agent used for requests to Reddit's API (required)
        :param str app_type: Type of the app. One of `web`, `installed`, or `script`
            (required)
        :param str redirect_uri: Redirect URI for Oauth2 flow. Defaults to user set
            redirect uri (required)
        :param str client_secret: Client secret of the Reddit App
        :param str app_description: Description of the Reddit App
        :param bool enabled: Allows the app to be used (defaults to `True`)
        :param Union[User,int,str] owner: Owner of the bot. Requires Admin to create for
            other users.

        :returns: RedditApp

        """
        data = {
            "app_name": name,
            "client_id": client_id,
            "user_agent": user_agent,
            "app_type": app_type,
            "redirect_uri": redirect_uri,
        }
        if client_secret:
            data["client_secret"] = client_secret
        if app_description:
            data["app_description"] = app_description
        if enabled:
            data["enabled"] = enabled
        if owner:
            data["owner_id"] = owner
        return _credmgr.post("/reddit_apps", data=data)

    def reddit(
        self,
        redditor=None,
        *,
        use_async=False,
        use_cache=True,
        reddit_class: Union[praw.Reddit, "asyncpraw.Reddit", None] = None,
        extra_reddit_kwargs=None,
    ) -> praw.Reddit:
        """Provide an optionally authenticated [Async] PRAW instance.

        :param str redditor: The redditor that you want the Reddit instance authorized
            as.
        :param bool use_async: Whether to return an asyncpraw instance.
        :param bool use_cache: Whether to fetch a new instance regardless if it was
            cached already.
        :param reddit_class: The Reddit class to use. Useful if you have a local version
            of PRAW.
        :param dict extra_reddit_kwargs: Extra kwargs to pass to the Reddit class.

        :returns: Reddit instance.

        """
        if extra_reddit_kwargs is None:
            extra_reddit_kwargs = {}
        if not (isinstance(self._reddit, praw.Reddit) == (not use_async)):
            self._reddit = None
        if not (use_cache and self._reddit):
            if not reddit_class:
                if use_async:
                    import asyncpraw

                    reddit_class = asyncpraw.Reddit
                else:
                    reddit_class = praw.Reddit
            if redditor:
                refresh_token = self._credmgr.refresh_token(redditor, self.id)
                if refresh_token:
                    extra_reddit_kwargs["refresh_token"] = refresh_token.refresh_token
            self._reddit = reddit_class(
                client_id=self.client_id,
                client_secret=self.client_secret,
                user_agent=self.user_agent,
                redirect_uri=self.redirect_uri,
                **extra_reddit_kwargs,
            )
        return self._reddit

    def gen_auth_url(self, scopes=None, permanent=False, user_verification=None):
        """Generate a URL for users to verify or authenticate their Reddit account.

        :param Union[list,str] scopes: List of scopes needed. Pass ``'all'`` for all
            scopes. The ``identity`` scope will always be included. (default:
            ``['identity']``)
        :param bool permanent: Determines if a refresh token will be provided. (default:
            ``False``)
        :param Union[UserVerification,int,str] user_verification: Link to a
                :class:`.UserVerification`
                object. Accepted values are:

                - An :class:`.UserVerification` object
                - An :class:`.UserVerification` ``id``
                - An ``user_id`` of a :class:`.UserVerification` record.

            If a :class:`.UserVerification` record does not exist, one will be created.

        :returns: Auth URL

        """
        from credmgr.models import UserVerification

        if scopes is None or scopes == "identity":
            scopes = ["identity"]
        elif scopes == "all":
            scopes = ["*"]
        if "identity" not in scopes and scopes != ["*"]:
            scopes = [scopes, "identity"]
        if permanent:
            duration = "permanent"
        else:
            duration = "temporary"
        u_verification = _resolve_model_from_input(
            self._credmgr, UserVerification, user_verification, "user_id"
        )
        if not u_verification and user_verification:
            u_verification = self._credmgr.user_verification.create(
                user_verification, self.id
            ).user_id
        if u_verification:
            state = base64.urlsafe_b64encode(
                f"{self.state}:{u_verification}".encode()
            ).decode()
        else:
            state = self.state
        return self.reddit().auth.url(scopes, state, duration)



"""Provide a decorator for deprecating camelCase arguments."""

import re
import warnings
from functools import wraps

regex = re.compile(r"([a-z])([A-Z])")


def camel_case_warning(attribute):
    """Warn about camelCase parameter names."""
    if regex.search(attribute):
        attribute = regex.sub(r"\1_\2", attribute).lower()
        warnings.warn(
            f"camelCase names have been deprecated. Please use {attribute} instead.",
            DeprecationWarning,
            stacklevel=3,
        )
    return attribute


def warn_camel_case(*items):
    """Warn about camelCase parameter names."""

    def wrapped(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if items:
                for item in items:
                    if item in list(kwargs.keys()):
                        new_name = camel_case_warning(item)
                        kwargs.setdefault(new_name, kwargs.pop(item))
            else:
                for item in list(kwargs.keys()):
                    new_name = camel_case_warning(item)
                    kwargs.setdefault(new_name, kwargs.pop(item))
            return func(*args, **kwargs)

        return wrapper

    return wrapped



"""Provide the CredentialManager class."""

import configparser
import os

from requests import Response

from . import models
from .auth import ApiTokenAuth
from .config import Config
from .deprecation_decorator import warn_camel_case
from .exceptions import InitializationError
from .models.utils import cached_property
from .requestor import Requestor, _urljoin

User = models.User
Bot = models.Bot
RedditApp = models.RedditApp
RefreshToken = models.RefreshToken
UserVerification = models.UserVerification
SentryToken = models.SentryToken
DatabaseCredential = models.DatabaseCredential


class CredentialManager:
    """The CredentialManager class provides a way to interact with CredentialManager's API.

    The official way to initialize an instance of this class is to:

    .. code-block:: python

        from credmgr import CredentialManager

        credential_manager = CredentialManager(api_token="LIqbGjAeep3Ws5DH3LOEQPmw8UZ6ek")

    .. note::

        It is recommended to use environment variables or a ``.credmgr.ini`` file. See
        :ref:`Auth documentation <auth>` for more details.

    """

    _default = None
    _endpoint = "/api/v1"

    def _objectify_response(self, response: Response):
        """Return a deserialized object from a response.

        :param response: The response to deserialize.

        :returns: A deserialized object.

        """
        data = response.json()
        if data:
            if isinstance(data, list):
                return [self._objectify(item) for item in data]
            return self._objectify(data)

    def _objectify(self, data):
        """Deserialize a dict or list into an object.

        :param data: Data to be deserialized.

        :returns: A credmgr object.

        """
        if data:
            resource_type = data.get("resource_type", "")
            model = getattr(models, resource_type, None)
            data = self._inject(data)
            return model.schema().load(data) if model else data

    def _inject(self, data: dict):
        """Injects the credential manager into the data.

        :param data: Data to be injected.

        :returns: The injected data.

        """
        if isinstance(data, dict):
            if "resource_type" in data:
                data["credmgr"] = self
            for key, value in list(data.items()):
                if isinstance(value, dict):
                    data[key] = self._inject(value)
                elif isinstance(value, list):
                    data[key] = list(map(self._inject, value))
        return data

    @warn_camel_case("configName", "sessionClass", "sessionKwargs", "apiToken")
    def __init__(
        self, config_name=None, session_class=None, session_kwargs=None, **kwargs
    ):
        """Initialize a CredentialManager instance.

        :param str config_name: The name of a section in your ``.credmgr.ini`` file that
            credmgr will load its configuration is loaded from. If ``config_name`` is
            ``None``, then it will look for it in the environment variable
            ``credmgr_config_name``. If it is not found there, the ``default`` section
            is used.
        :param Session session_class: A Session class that will be used to create a
            requestor. If not set, use ``requests.Session`` (default: None).
        :param dict session_kwargs: Dictionary with additional keyword arguments used to
            initialize the session (default: None).

        Additional keyword arguments will be used to initialize the :class:`.Config`
        object. This can be used to specify configuration settings during instantiation
        of the :class:`.CredentialManager` instance. For more details please see
        :ref:`configuration`.

        Required settings are:

        - api_token

        OR

        - username
        - password

        .. warning::

            Using an API Token instead of a username/password is strongly recommended!

        The ``session_class`` and ``session_kwargs`` allow for customization of the
        session :class:`.CredentialManager` will use. This allows, e.g., easily adding
        behavior to the requestor or wrapping its |Session|_ in a caching layer. Example
        usage:

        .. |Session| replace:: ``Session``

        .. _session: https://2.python-requests.org/en/master/api/#requests.Session

        .. code-block:: python

            import json, betamax, requests


            class JSONDebugRequestor(Requestor):
                def request(self, *args, **kwargs):
                    response = super().request(*args, **kwargs)
                    print(json.dumps(response.json(), indent=4))
                    return response


            my_session = betamax.Betamax(requests.Session())
            credential_manager = CredentialManager(
                ..., session_class=JSONDebugRequestor, session=my_session
            )

        """
        if session_kwargs is None:
            session_kwargs = {}
        config_section = None
        try:
            config_section = (
                config_name or os.getenv("credmgr_config_name") or "DEFAULT"
            )
            self.config = Config(config_section, **kwargs)
        except configparser.NoSectionError as exc:
            if config_section is not None:
                exc.message += "\nYou provided the name of a .credmgr.ini section that doesn't exist."
            raise

        self._server = _urljoin(self.config.server, self.config.endpoint)
        api_token = self.config.api_token
        username = self.config.username
        password = self.config.password

        init_error_message = "Required settings are missing. Either 'api_token' OR 'username' and 'password' must be specified, These settings can be provided in a .credmgr.ini file, as a keyword argument during the initialization of the `CredentialManager` class, or as an environment variable."
        if all([api_token, username, password]):
            raise InitializationError(init_error_message)
        if api_token:
            self._auth = ApiTokenAuth(api_token)
        elif username and password:
            self._auth = (username, password)
        else:
            raise InitializationError(
                "API Token or a username/password pair must be set."
            )

        self._requestor = Requestor(
            self._server, self._auth, session_class, **session_kwargs
        )
        self._current_user = None
        self._user_defaults = None
        self.get_user_default = lambda key, default: self.user_defaults.get(
            key, default
        )
        self.user = models.UserHelper(self)
        """An instance of :class:`.UserHelper`.

        Provides an interface for interacting with :class:`.User`. For example, to get a
        :class:`.User` with :attr:`id` of ``1`` you can do:

        .. code-block:: python

            user = credential_manager.user(1)
            print(user.id)

        To create a :class:`.User` do:

        .. code-block:: python

            user = credential_manager.user.create(**user_kwargs)

        See :meth:`~.UserHelper.create` for the required params.

        """
        self.bot = models.BotHelper(self)
        """An instance of :class:`.BotHelper`.

        Provides an interface for interacting with :class:`.Bot`. For example, to get a
        :class:`.Bot` with :attr:`id` of ``1`` you can do:

        .. code-block:: python

            bot = credential_manager.bot(1)
            print(bot.id)

        To create a :class:`.Bot` do:

        .. code-block:: python

            bot = credential_manager.bot.create(**bot_kwargs)

        See :meth:`~.BotHelper.create` for the required params.

        """
        self.reddit_app = models.RedditAppHelper(self)
        """An instance of :class:`.RedditAppHelper`.

        Provides an interface for interacting with :class:`.RedditApp`. For example, to
        get a :class:`.RedditApp` with :attr:`id` of ``1`` you can do:

        .. code-block:: python

            reddit_app = credential_manager.reddit_app(1)
            print(reddit_app.id)

        To create a :class:`.RedditApp` do:

        .. code-block:: python

            reddit_app = credential_manager.reddit_app.create(**reddit_app_kwargs)

        See :meth:`~.RedditAppHelper.create` for the required params.

        """
        self.refresh_token = models.RefreshTokenHelper(self)
        """An instance of :class:`.RefreshTokenHelper`.

        Provides an interface for interacting with :class:`.RefreshToken`. For example
        to get a :class:`.RefreshToken` with :attr:`id` of ``1`` you can do:

        .. code-block:: python

            refresh_token = credential_manager.refresh_token(1)
            print(refresh_token.id)

        .. note::

            Refresh tokens cannot be manually created.

        """
        self.user_verification = models.UserVerificationHelper(self)
        """An instance of :class:`.UserVerificationHelper`.

        Provides an interface for interacting with :class:`.UserVerification`. For
        example to get a :class:`.UserVerification` with :attr:`id` of ``1`` you can do:

        .. code-block:: python

            user_verification = credential_manager.user_verification(1)
            print(user_verification.id)

        To create a :class:`.UserVerification` do:

        .. code-block:: python

            user_verification = credential_manager.user_verification.create(
                **user_verification_kwargs
            )

        See :meth:`~.UserVerificationHelper.create` for the required params.

        """
        self.sentry_token = models.SentryTokenHelper(self)
        """An instance of :class:`.SentryTokenHelper`.

        Provides an interface for interacting with :class:`.SentryToken`. For example to
        get a :class:`.SentryToken` with :attr:`id` of ``1`` you can do:

        .. code-block:: python

            sentry_token = credential_manager.sentry_token(1)
            print(sentry_token.id)

        To create a :class:`.SentryToken` do:

        .. code-block:: python

            sentry_token = credential_manager.sentry_token.create(**sentry_token_kwargs)

        See :meth:`~.SentryTokenHelper.create` for the required params.

        """
        self.database_credential = models.DatabaseCredentialHelper(self)
        """An instance of :class:`.DatabaseCredentialHelper`.

        Provides an interface for interacting with :class:`.DatabaseCredential`. For
        example to get a :class:`.DatabaseCredential` with :attr:`id` of ``1`` you can
        do:

        .. code-block:: python

            database_credential = credential_manager.database_credential(1)
            print(database_credential.id)

        To create a :class:`.DatabaseCredential` do:

        .. code-block:: python

            database_credential = credential_manager.database_credential.create(
                **database_credential_kwargs
            )

        See :meth:`~.DatabaseCredentialHelper.create` for the required params.

        """

    def users(self, batch_size=10, limit=None):
        """List Users.

        :param int batch_size: Number of Users to return in each batch (default: ``20``)
        :param int limit: Maximum number of Users to return

        :returns: A :class:`.Paginator` to iterate through the Users

        """
        return User(self).list_items(batch_size=batch_size, limit=limit)

    def bots(self, batch_size=20, limit=None, owner=None):
        """List Bots.

        :param int batch_size: Number of Bots to return in each batch (default: ``20``)
        :param int limit: Maximum number of Bots to return
        :param Union[int,str,User] owner: Return Bots that are owner by this user

        :returns: A :class:`.Paginator` to iterate through the Bots

        """
        return Bot(self).list_items(batch_size=batch_size, limit=limit, owner=owner)

    def reddit_apps(self, batch_size=20, limit=None, owner=None):
        """List RedditApps.

        :param int batch_size: Number of RedditApps to return in each batch (default:
            ``20``)
        :param int limit: Maximum number of RedditApps to return
        :param Union[int,str,User] owner: Return RedditApps that are owner by this user

        :returns: A :class:`.Paginator` to iterate through the Reddit Apps

        """
        return RedditApp(self).list_items(
            batch_size=batch_size, limit=limit, owner=owner
        )

    def refresh_tokens(self, batch_size=20, limit=None, owner=None):
        """List RefreshTokens.

        :param int batch_size: Number of RefreshTokens to return in each batch (default:
            ``20``)
        :param int limit: Maximum number of RefreshTokens to return
        :param Union[int,str,User] owner: Return RefreshTokens that are owned by this
            user

        :returns: A :class:`.Paginator` to iterate through the Refresh Tokens

        .. note::

            This is *not* the intended way to fetch refresh tokens. See:
            :meth:`~.RedditApp.reddit` for obtaining an authorized reddit instance.

        """
        return RefreshToken(self).list_items(
            batch_size=batch_size, limit=limit, owner=owner
        )

    def user_verifications(self, batch_size=20, limit=None, owner=None):
        """List UserVerifications.

        :param int batch_size: Number of UserVerifications to return in each batch
            (default: ``20``)
        :param int limit: Maximum number of UserVerifications to return
        :param Union[int,str,User] owner: Return UserVerifications that are owned by
            this user

        :returns: A :class:`.Paginator` to iterate through the UserVerifications.

        """
        return UserVerification(self).list_items(
            batch_size=batch_size, limit=limit, owner=owner
        )

    def sentry_tokens(self, batch_size=20, limit=None, owner=None):
        """List SentryTokens.

        :param int batch_size: Number of SentryTokens to return in each batch (default:
            ``20``)
        :param int limit: Maximum number of SentryTokens to return
        :param Union[int,str,User] owner: Return SentryTokens that are owned by this
            user

        :returns: A :class:`.Paginator` to iterate through the SentryTokens

        """
        return SentryToken(self).list_items(
            batch_size=batch_size, limit=limit, owner=owner
        )

    def database_credentials(self, batch_size=20, limit=None, owner=None):
        """List DatabaseCredentials.

        :param int batch_size: Number of DatabaseCredentials to return in each batch
            (default: ``20``)
        :param int limit: Maximum number of DatabaseCredentials to return
        :param Union[int,str,User] owner: Return DatabaseCredentials that are owned by
            this user

        :returns: A :class:`.Paginator` to iterate through the DatabaseCredentials

        """
        return DatabaseCredential(self).list_items(
            batch_size=batch_size, limit=limit, owner=owner
        )

    @cached_property
    def current_user(self) -> User:
        """Get the currently authenticated :class:`.User`."""
        if not self._current_user:
            self._current_user = self.get("/users/me")
        return self._current_user

    @cached_property
    def user_defaults(self):
        """Get the currently authenticated :class:`.User`'s default settings."""
        if not self._user_defaults:
            self._user_defaults = self.current_user.default_settings
        return self._user_defaults

    def get(self, path, params=None):
        """Return parsed objects returned from a GET request to ``path``.

        :param path: The path to fetch.
        :param params: The query parameters to add to the request (default: None).

        """
        return self._objectify_response(
            self._requestor.request(path, "GET", params=params)
        )

    def post(self, path, data):
        """Return parsed objects returned from a POST request to ``path``.

        :param path: The path to fetch.
        :param data: Dictionary, bytes, or file-like object to send in the body of the
            request (default: None).

        """
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        return self._objectify_response(
            self._requestor.request(path, "POST", data=data, headers=headers)
        )

    def patch(self, path, data):
        """Return parsed objects returned from a PATCH request to ``path``.

        :param path: The path to fetch.
        :param data: Dictionary, bytes, or file-like object to send in the body of the
            request (default: None).

        """
        return self._objectify_response(
            self._requestor.request(path, "PATCH", json=data)
        )

    def delete(self, path):
        """Return parsed objects returned from a DELETE request to ``path``.

        :param path: The path to fetch.

        """
        self._requestor.request(path, "DELETE")



"""Provide the RefreshToken class."""

from ..mixins import BaseApp, RedditAppMixin


class RefreshToken(BaseApp, RedditAppMixin):
    """A class for RefreshToken instances."""

    _can_fetch_by_name = True
    _credmgr_callable = "refresh_token"
    _editable_attrs = ["refresh_token"]
    _get_by_name_attrs = ["redditor", "reddit_app_id"]
    _get_by_name_path = "by_redditor"
    _name_attr = "redditor"
    _path = "/refresh_tokens"

    @classmethod
    def schema(cls):
        """Return the schema for this model."""
        if cls._schema is None:
            from .schemas import RefreshTokenSchema

            cls._schema = RefreshTokenSchema()
        return cls._schema

    def __init__(self, credmgr, **kwargs):
        """Initialize a Refresh Token instance.

        Refresh Tokens are for authenticating with Reddit as a Reddit that has
        authorized a `.RedditApp` to access their Reddit account.

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param int id: ID of this Refresh Token.
        :param str redditor: Redditor this Refresh Token is for.
        :param str reddit_app_id: ID of the `.RedditApp` this Refresh Token is for.
        :param str refresh_token: The Refresh Token to pass to a Reddit instance.
        :param list[str] scopes: The OAuth2 scopes this Refresh Token grants access to.
        :param int owner_id: ID of the `.User` that owns this Refresh Token.
        :param datetime.datetime issued_at: Date and time this Refresh Token was issued.
        :param bool revoked: Indicates if this Refresh Token was revoked or superseded
            by another Refresh Token.
        :param datetime.datetime revoked_at: Date and time this Refresh Token was
            revoked or superseded.

        """
        super().__init__(credmgr, **kwargs)



"""Provide CredentialManager schemas."""

from marshmallow import Schema, post_load
from marshmallow.fields import (
    AwareDateTime,
    Boolean,
    Dict,
    Field,
    Integer,
    List,
    Nested,
    Raw,
    String,
)

from . import (
    Bot,
    DatabaseCredential,
    RedditApp,
    RefreshToken,
    SentryToken,
    User,
    UserVerification,
)


class PortField(Field):
    def _serialize(self, value, *args, **kwargs):
        if value is None:
            return None
        return int(value)

    def _deserialize(self, value, *args, **kwargs):
        return str(value)


class BaseSchema(Schema):
    """Base schema for all schemas."""

    _model = None
    credmgr = Raw(load_only=True)
    id = Integer()
    name = String(data_key="app_name")
    resource_type = String(load_only=True)

    def get_attribute(self, obj, attr, default):  # pragma: no cover
        """Get an attribute from an object. Used for dumping."""
        return super().get_attribute(obj, f"__dict__.{attr}", default)

    @post_load
    def objectify(self, data, **_):
        """Return an object if the schema's specified model."""
        return self._model(**data)


class BaseAppSchema(BaseSchema):
    """Base schema for all app schemas."""

    enabled = Boolean()
    owner_id = Integer()


class RedditAppSchema(BaseAppSchema):
    """Schema for :class:`.RedditApp`."""

    _model = RedditApp
    app_description = String(allow_none=True)
    app_type = String()
    client_id = String()
    client_secret = String(allow_none=True)
    redirect_uri = String()
    state = String()
    user_agent = String()


class DatabaseCredentialSchema(BaseAppSchema):
    """Schema for :class:`.DatabaseCredential`."""

    _model = DatabaseCredential
    database_username = String()
    database_host = String()
    database = String()
    database_flavor = String()
    database_port = PortField()
    database_password = String(allow_none=True)
    use_ssh = Boolean()
    ssh_host = String(allow_none=True)
    ssh_port = PortField(allow_none=True)
    ssh_username = String(allow_none=True)
    ssh_password = String(allow_none=True)
    use_ssh_key = Boolean()
    private_key = String(allow_none=True)
    private_key_passphrase = String(allow_none=True)


class SentryTokenSchema(BaseAppSchema):
    """Schema for :class:`.SentryToken`."""

    _model = SentryToken
    dsn = String()


class UserVerificationSchema(BaseAppSchema):
    """Schema for :class:`.UserVerification`."""

    _model = UserVerification
    extra_data = Dict(keys=String(), allow_none=True)
    reddit_app_id = Integer(allow_none=True)
    redditor = String(allow_none=True)
    user_id = String()


class RefreshTokenSchema(BaseSchema):
    """Schema for :class:`.RefreshToken`."""

    _model = RefreshToken
    issued_at = AwareDateTime()
    owner_id = Integer()
    reddit_app_id = Integer()
    redditor = String()
    refresh_token = String()
    revoked = Boolean()
    revoked_at = AwareDateTime(allow_none=True)
    scopes = List(String())


class BotSchema(BaseAppSchema):
    """Schema for :class:`.Bot`."""

    _model = Bot
    database_credential = Nested(DatabaseCredentialSchema, allow_none=True)
    reddit_app = Nested(RedditAppSchema, allow_none=True)
    sentry_token = Nested(SentryTokenSchema, allow_none=True)


class UserSchema(BaseSchema):
    """Schema for :class:`.User`."""

    _model = User
    bots = List(Nested(BotSchema))
    created = AwareDateTime()
    database_credentials = List(Nested(DatabaseCredentialSchema))
    default_settings = Dict()
    is_active = Boolean()
    is_admin = Boolean()
    is_regular_user = Boolean()
    reddit_apps = List(Nested(RedditAppSchema))
    reddit_username = String(allow_none=True)
    sentry_tokens = List(Nested(SentryTokenSchema))
    updated = AwareDateTime()
    username = String()



from credmgr.models import RedditApp, UserVerification


def test_resolve_user_with_object(recorder, credential_manager):
    user = credential_manager.current_user
    reddit_apps = credential_manager.reddit_apps(owner=user)
    for reddit_app in reddit_apps:
        assert isinstance(reddit_app, RedditApp)
        assert reddit_app.owner == user


def test_resolve_user_with_username(recorder, credential_manager):
    user = credential_manager.current_user
    reddit_apps = credential_manager.reddit_apps(owner=user.username)
    for reddit_app in reddit_apps:
        assert isinstance(reddit_app, RedditApp)
        assert reddit_app.owner == user


def test_resolve_model_with_object(recorder, credential_manager):
    reddit_app = credential_manager.reddit_app(26)
    user_verification = credential_manager.user_verification.create(
        user_id="testuserid", reddit_app=reddit_app
    )
    assert isinstance(user_verification, UserVerification)


def test_resolve_model_with_id(recorder, credential_manager):
    reddit_app = credential_manager.reddit_app(26)
    user_verification = credential_manager.user_verification.create(
        user_id="testuserid", reddit_app=reddit_app.id
    )
    assert isinstance(user_verification, UserVerification)



"""Auth base."""

from requests.auth import AuthBase


class ApiTokenAuth(AuthBase):
    """Class for generating API Token request headers."""

    def __init__(self, api_token):
        """Initialize the :class:`ApiTokenAuth` class."""
        self.api_token = api_token

    def __call__(self, request):
        """Generate the request headers."""
        request.headers["X-API-TOKEN"] = self.api_token
        return request



"""Provides the code to load credmgr's configuration file `~/.credmgr.ini`."""

import configparser
import os
import sys
import warnings
from pathlib import Path
from threading import Lock


class Config:
    """Class for loading credmgr's config."""

    _config = None
    _lock = Lock()

    @staticmethod
    def _append_file_name(dirname):
        return os.path.join(dirname, ".credmgr.ini")

    @classmethod
    def _load_config(cls):
        """Attempt to load settings from various .credmgr.ini files."""
        config = configparser.ConfigParser()
        root_dir = os.path.dirname(sys.modules[__name__].__file__)
        locations = [
            cls._append_file_name(root_dir),
            cls._append_file_name(Path.home()),
            ".credmgr.ini",
        ]
        config.read(locations)
        cls._config = config

    def __init__(self, config_name, **kwargs):
        """Initialize a Config instance."""
        with Config._lock:
            if Config._config is None:
                self._load_config()
        self._settings = kwargs
        self.custom_settings = dict(
            Config._config.items(config_name),
            **{k.lower(): v for k, v in kwargs.items() if v},
        )
        self.server = self.api_token = self.username = self.password = None
        self._initialize_config()

    def _fetch(self, key, default=None):
        config_file_value = self.custom_settings.pop(key.lower(), default)
        env_override = os.getenv(f"credmgr_{key}")
        return env_override or config_file_value

    def _initialize_config(self):
        self.server = self._fetch("server")
        self.endpoint = self._fetch("endpoint")
        old_token = self._fetch("apitoken")
        new_token = self._fetch("api_token")
        if old_token and not new_token:  # pragma: no cover
            warnings.warn(
                "apitoken is deprecated, use api_token instead.",
                DeprecationWarning,
                stacklevel=3,
            )
        self.api_token = new_token or old_token
        self.username = self._fetch("username")
        self.password = self._fetch("password")



import pytest

from credmgr.exceptions import Conflict, NotFound, ServerError
from credmgr.models import DatabaseCredential

data = {
    "name": "test_database_credential",
    "database_username": "database_username",
    "database_host": "database_host",
    "database": "database",
    "database_flavor": "database_flavor",
    "database_port": "6543",
    "database_password": "database_password",
    "use_ssh": True,
    "ssh_host": "ssh_host",
    "ssh_port": "1234",
    "ssh_username": "ssh_username",
    "ssh_password": "ssh_password",
    "use_ssh_key": True,
    "private_key": "private_key",
    "private_key_passphrase": "private_key_passphrase",
}


def test_create_database_credential(recorder, credential_manager):
    database_credential = credential_manager.database_credential.create(**data)
    print(database_credential.id)
    for key, value in data.items():
        assert getattr(database_credential, key) == value


def test_create_database_credential_other_user(recorder, credential_manager):
    new_data = {**data, "owner": 4}
    database_credential = credential_manager.database_credential.create(
        **{**new_data, "owner": 4}
    )
    for key, value in new_data.items():
        if key == "owner":
            assert getattr(database_credential, key).id == value
            continue
        assert getattr(database_credential, key) == value


def test_create_database_credential_bad_params(recorder, credential_manager):
    data = {"name": "se"}
    with pytest.raises(ServerError):
        _ = credential_manager.database_credential.create(**data)


def test_create_database_credential_existing(recorder, credential_manager):
    with pytest.raises(Conflict):
        _ = credential_manager.database_credential.create(**data)


def test_delete_database_credential(recorder, credential_manager):
    database_credential = credential_manager.database_credential(3)
    database_credential.delete()
    with pytest.raises(NotFound):
        _ = credential_manager.database_credential(3)


def test_edit_database_credential(recorder, credential_manager):
    database_credential = credential_manager.database_credential(3)
    database_credential.edit(database_username="new_username")
    assert database_credential.database_username == "new_username"


def test_edit_database_credential_conflicting_data(recorder, credential_manager):
    database_credential = credential_manager.database_credential(4)
    with pytest.raises(Conflict):
        database_credential.edit(name="test_database_credential")


def test_list_database_credentials(recorder, credential_manager):
    database_credentials = credential_manager.database_credentials()
    for database_credential in database_credentials:
        assert isinstance(database_credential, DatabaseCredential)



import pytest
from betamax import Betamax

from credmgr.exceptions import Conflict, InitializationError, NotFound
from credmgr.models import DatabaseCredential, RedditApp, SentryToken, User

data = {
    "username": "username",
    "password": "password",
    "reddit_username": "reddit_username",
    "default_settings": {"database_flavor": "postgres", "user_agent": "user_agent"},
}


def test_create_user(recorder, credential_manager):
    data = {
        "username": "username",
        "password": "password",
        "reddit_username": "reddit_username",
        "default_settings": {"database_flavor": "postgres", "user_agent": "user_agent"},
    }
    user = credential_manager.user.create(**data)
    data.pop("password")
    for key, value in data.items():
        assert getattr(user, key) == value


def test_create_user_bad_params(recorder, credential_manager):
    data = {"username": "te"}
    with pytest.raises(TypeError):
        _ = credential_manager.user.create(**data)


def test_create_user_existing(recorder, credential_manager):
    data = {"username": "username", "password": "password"}
    with pytest.raises(Conflict):
        _ = credential_manager.user.create(**data)


def test_delete_user(recorder, credential_manager):
    user = credential_manager.user(12)
    user.delete()
    with pytest.raises(NotFound):
        _ = credential_manager.user(12)


def test_edit_user(recorder, credential_manager):
    user = credential_manager.user("test2")
    data = {
        "username": "new_username",
        "reddit_username": "new_reddit_username",
        "default_settings": {
            "database_host": "localhost",
            "database_flavor": "postgres",
            "user_agent": "user_agent",
        },
    }
    user.edit(**data)
    modified_user = credential_manager.user("new_username")
    for key, value in data.items():
        assert getattr(modified_user, key) == value


def test_edit_user_conflicting_data(recorder, credential_manager):
    user = credential_manager.user(6)
    with pytest.raises(Conflict):
        user.edit(username="username")


def test_list_users(recorder, credential_manager):
    users = credential_manager.users()
    for user in users:
        assert isinstance(user, User)


def test_list_users_with_user(recorder, credential_manager):
    with pytest.raises(InitializationError):
        _ = credential_manager.user()


def test_apps(recorder, credential_manager):
    user = credential_manager.current_user
    all_apps = user.apps()
    assert isinstance(all_apps, dict)
    for apps in all_apps.values():
        for app in apps:
            assert isinstance(app, (DatabaseCredential, RedditApp, SentryToken))


def test_apps_only(credential_manager):
    with Betamax(credential_manager._requestor._session).use_cassette(
        "TestUser.test_apps"
    ):
        user = credential_manager.current_user
        sentry_tokens = user.apps("sentry_tokens")
        assert isinstance(sentry_tokens, list)
        for token in sentry_tokens:
            assert isinstance(token, SentryToken)


def test_apps_only_invalid(credential_manager):
    with Betamax(credential_manager._requestor._session).use_cassette(
        "TestUser.test_apps"
    ):
        user = credential_manager.current_user
        with pytest.raises(InitializationError):
            _ = user.apps("invalid")



"""Provide the BaseApp class."""

from .deletable import DeletableMixin
from .editable import EditableMixin
from .owner import OwnerMixin


class BaseApp(DeletableMixin, EditableMixin, OwnerMixin):
    """The base class for all app classes."""

    _deserialize_mapping = {"app_name": "name"}
    _editable_attrs = ["name"]
    _name_attr = "name"
    _serialize_mapping = {"name": "app_name"}



"""Provide the RedditAppMixin class."""

from ..models.utils import cached_property
from . import BaseModel


class RedditAppMixin(BaseModel):
    """Interface for classes that have an associated :class:`RedditApp`."""

    _editable_attrs = []

    @cached_property
    def reddit_app(self):
        """Return the associated :class:`RedditApp`."""
        if not self._fetched:
            self._fetch()
        reddit_app = self._credmgr.reddit_app(id=self.reddit_app_id)
        return reddit_app



import time
from functools import partial
from unittest import mock

import asyncpraw
import praw
import pytest

from credmgr.exceptions import Conflict, InitializationError, NotFound, ServerError
from credmgr.models import RedditApp


def test_create_reddit_app(recorder, credential_manager):
    data = {
        "name": "test_reddit_app",
        "client_id": "client_id",
        "client_secret": "client_secret",
        "user_agent": "user_agent",
        "app_description": "app_description",
    }
    reddit_app = credential_manager.reddit_app.create(**data)
    for key, value in data.items():
        assert getattr(reddit_app, key) == value


def test_create_reddit_app_other_user(recorder, credential_manager):
    data = {"name": "test_reddit_app", "client_id": "client_id", "owner": 4}
    reddit_app = credential_manager.reddit_app.create(**data)
    for key, value in data.items():
        if key == "owner":
            assert getattr(reddit_app, key).id == value
            continue
        assert getattr(reddit_app, key) == value


def test_create_reddit_app_bad_params(recorder, credential_manager):
    data = {"name": "te", "client_id": "client_id_2"}
    with pytest.raises(ServerError):
        _ = credential_manager.reddit_app.create(**data)


def test_create_reddit_app_existing(recorder, credential_manager):
    data = {"name": "reddit_app", "client_id": "client_id"}
    with pytest.raises(Conflict):
        _ = credential_manager.reddit_app.create(**data)


def test_delete_reddit_app(recorder, credential_manager):
    reddit_app = credential_manager.reddit_app(3)
    reddit_app.delete()
    with pytest.raises(NotFound):
        _ = credential_manager.reddit_app(3)


def test_edit_reddit_app(recorder, credential_manager):
    reddit_app = credential_manager.reddit_app(6)
    reddit_app.edit(client_id="new_client_id")
    assert reddit_app.client_id == "new_client_id"


def test_edit_reddit_app_conflicting_data(recorder, credential_manager):
    reddit_app = credential_manager.reddit_app(6)
    with pytest.raises(Conflict):
        reddit_app.edit(client_id="client_id")


def test_list_reddit_apps(recorder, credential_manager):
    reddit_apps = credential_manager.reddit_apps()
    for reddit_app in reddit_apps:
        assert isinstance(reddit_app, RedditApp)


def test_list_reddit_apps_with_reddit_app(recorder, credential_manager):
    with pytest.raises(InitializationError):
        _ = credential_manager.reddit_app()


def test_create_reddit_app_without_user_agent(recorder, credential_manager):
    data = {
        "name": "test_reddit_app",
        "client_id": "client_id",
        "client_secret": "client_secret",
        "app_description": "app_description",
    }
    reddit_app = credential_manager.reddit_app.create(**data)
    assert reddit_app.user_agent == "python:test_reddit_app by u/Lil_SpazJoekp"
    for key, value in data.items():
        assert getattr(reddit_app, key) == value


class CustomSyncReddit(praw.Reddit):
    pass


class CustomAsyncReddit(asyncpraw.Reddit):
    pass


@pytest.mark.parametrize("use_async", [True, False], ids=["use_async", "not_use_async"])
@pytest.mark.parametrize("use_cache", [True, False], ids=["use_cache", "not_use_cache"])
@pytest.mark.parametrize(
    "use_custom_reddit_class",
    [True, False],
    ids=["custom_reddit_class", "no_custom_class"],
)
def test_reddit_app_reddit(
    recorder, credential_manager, use_async, use_cache, use_custom_reddit_class
):
    reddit_app = credential_manager.reddit_app(2)
    reddit_class = None
    if use_custom_reddit_class:
        reddit_class = CustomAsyncReddit if use_async else CustomSyncReddit
    reddit = reddit_app.reddit(use_async=use_async, reddit_class=reddit_class)
    same = reddit is reddit_app.reddit(
        use_async=use_async, use_cache=use_cache, reddit_class=reddit_class
    )
    assert (
        type(reddit) is (CustomAsyncReddit if use_async else CustomSyncReddit)
        if reddit_class
        else (asyncpraw.Reddit if use_async else praw.Reddit)
    )
    assert same == use_cache
    assert isinstance(reddit, praw.Reddit if not use_async else asyncpraw.Reddit)


def patched__request_token(self, **data):
    payload = {
        "access_token": "access_token",
        "token_type": "bearer",
        "expires_in": 3600,
        "refresh_token": "new_token",
        "scope": "account creddits edit flair history identity livemanage modconfig modcontributors modflair modlog modmail modothers modposts modself modtraffic modwiki mysubreddits privatemessages read report save structuredstyles submit subscribe vote wikiedit wikiread",
    }
    pre_request_time = time.time()
    self._expiration_timestamp = pre_request_time - 10 + payload["expires_in"]
    self.access_token = payload["access_token"]
    if "refresh_token" in payload:
        self.refresh_token = payload["refresh_token"]
    self.scopes = set(payload["scope"].split(" "))


def test_reddit_app_reddit_with_redditor(recorder, credential_manager):
    reddit_app = credential_manager.reddit_app(2)
    reddit = reddit_app.reddit("Lil_SpazJoekp")
    assert isinstance(reddit, praw.Reddit)
    assert not reddit.read_only
    response = mock.Mock(status_code=200)
    response.json.return_value = {"name": "Lil_SpazJoekp"}
    with mock.patch(
        "prawcore.requestor.Requestor.request", return_value=response, create=True
    ):
        with mock.patch("prawcore.rate_limit.RateLimiter.delay"):
            with mock.patch("prawcore.rate_limit.RateLimiter.update"):
                reddit._core._authorizer._request_token = partial(
                    patched__request_token, self=reddit._core._authorizer
                )
                reddit._core._authorizer.refresh()
                assert reddit.user.me() == "Lil_SpazJoekp"


def test_reddit_app_reddit_with_bad_redditor(recorder, credential_manager):
    reddit_app = credential_manager.reddit_app(2)
    with pytest.raises(NotFound):
        _ = reddit_app.reddit("BadRedditor")


def test_reddit_app_gen_auth_url_verification_scopes(recorder, credential_manager):
    reddit_app = credential_manager.reddit_app(19)
    url = reddit_app.gen_auth_url(scopes="read")
    assert (
        url
        == "https://www.reddit.com/api/v1/authorize?client_id=client_id&duration=temporary&redirect_uri=https%3A%2F%2Fcredmgr.jesassn.org%2Foauth2%2Freddit_callback&response_type=code&scope=read+identity&state=91893362b02d01c00fb71bc6be02065dedc3356595fc648581ab59a94864a7e4"
    )


def test_reddit_app_gen_auth_url_verification_all_scopes(recorder, credential_manager):
    reddit_app = credential_manager.reddit_app(19)
    url = reddit_app.gen_auth_url(scopes="all")
    assert (
        url
        == "https://www.reddit.com/api/v1/authorize?client_id=client_id&duration=temporary&redirect_uri=https%3A%2F%2Fcredmgr.jesassn.org%2Foauth2%2Freddit_callback&response_type=code&scope=%2A&state=91893362b02d01c00fb71bc6be02065dedc3356595fc648581ab59a94864a7e4"
    )


def test_reddit_app_gen_auth_url_verification(recorder, credential_manager):
    reddit_app = credential_manager.reddit_app(6)
    url = reddit_app.gen_auth_url()
    assert (
        url
        == "https://www.reddit.com/api/v1/authorize?client_id=new_client_id&duration=temporary&redirect_uri=https%3A%2F%2Fcredmgr.jesassn.org%2Foauth2%2Freddit_callback&response_type=code&scope=identity&state=7ce31a95e1abc2a41673406ff87da51932acf1047cbf2eec1beb1f783d81cfd4"
    )


def test_reddit_app_gen_auth_url_authencation(recorder, credential_manager):
    reddit_app = credential_manager.reddit_app(6)
    url = reddit_app.gen_auth_url(permanent=True)
    assert (
        url
        == "https://www.reddit.com/api/v1/authorize?client_id=new_client_id&duration=permanent&redirect_uri=https%3A%2F%2Fcredmgr.jesassn.org%2Foauth2%2Freddit_callback&response_type=code&scope=identity&state=7ce31a95e1abc2a41673406ff87da51932acf1047cbf2eec1beb1f783d81cfd4"
    )


def test_reddit_app_gen_auth_url_user_verification(recorder, credential_manager):
    reddit_app = credential_manager.reddit_app(6)
    url = reddit_app.gen_auth_url(user_verification=22)
    assert (
        url
        == "https://www.reddit.com/api/v1/authorize?client_id=new_client_id&duration=temporary&redirect_uri=https%3A%2F%2Fcredmgr.jesassn.org%2Foauth2%2Freddit_callback&response_type=code&scope=identity&state=N2NlMzFhOTVlMWFiYzJhNDE2NzM0MDZmZjg3ZGE1MTkzMmFjZjEwNDdjYmYyZWVjMWJlYjFmNzgzZDgxY2ZkNDoyMg%3D%3D"
    )


def test_reddit_app_gen_auth_url_user_verification_create(recorder, credential_manager):
    reddit_app = credential_manager.reddit_app(19)
    url = reddit_app.gen_auth_url(user_verification="new_id")
    assert (
        url
        == "https://www.reddit.com/api/v1/authorize?client_id=client_id&duration=temporary&redirect_uri=https%3A%2F%2Fcredmgr.jesassn.org%2Foauth2%2Freddit_callback&response_type=code&scope=identity&state=OTE4OTMzNjJiMDJkMDFjMDBmYjcxYmM2YmUwMjA2NWRlZGMzMzU2NTk1ZmM2NDg1ODFhYjU5YTk0ODY0YTdlNDpuZXdfaWQ%3D"
    )


def test_reddit_app_gen_auth_url_user_verification_create_existing(
    recorder, credential_manager
):
    reddit_app = credential_manager.reddit_app(22)
    url = reddit_app.gen_auth_url(user_verification="new_id")
    assert (
        url
        == "https://www.reddit.com/api/v1/authorize?client_id=client_id&duration=temporary&redirect_uri=https%3A%2F%2Fcredmgr.jesassn.org%2Foauth2%2Freddit_callback&response_type=code&scope=identity&state=OTE4OTMzNjJiMDJkMDFjMDBmYjcxYmM2YmUwMjA2NWRlZGMzMzU2NTk1ZmM2NDg1ODFhYjU5YTk0ODY0YTdlNDpuZXdfaWQ%3D"
    )



"""Provide the BaseModel class."""

from typing import TYPE_CHECKING

from ..deprecation_decorator import camel_case_warning, warn_camel_case
from ..exceptions import InitializationError
from ..models.utils import _resolve_user

if TYPE_CHECKING:  # pragma: no cover
    from credmgr import CredentialManager


class BaseModel:
    """Superclass for all models in credmgr."""

    _serialize_mapping = {}
    _attr_types = {}
    _can_fetch_by_name = False
    _credmgr_callable = None
    _get_by_name_path = "by_name"
    _name_attr = "name"
    _get_by_name_attrs = []  # falls back to _name_attr if empty
    _path = None
    _schema = None

    @warn_camel_case()
    def __init__(self, credmgr: "CredentialManager", **kwargs):
        """Initialize a BaseModel instance.

        :param credmgr: An instance of :class:`.CredentialManager`.

        """
        self._credmgr = credmgr
        if kwargs:
            for attribute, value in kwargs.items():
                setattr(self, attribute, value)
        self._fetched = False

    def __getattr__(self, attribute):  # pragma: no cover
        """Return the value of ``attribute``."""
        attribute = camel_case_warning(attribute)
        if attribute in self.__dict__:
            return self.__dict__[attribute]
        if not attribute.startswith("_") and not self._fetched:
            self._fetch()
            return getattr(self, attribute)
        raise AttributeError(
            f"{self.__class__.__name__!r} object has no attribute {attribute!r}"
        )

    def _get(self, id):
        self.__dict__ = self._credmgr.get(f"{self._path}/{id}").__dict__

    def _get_by_name(self):
        data = {}
        if not self._get_by_name_attrs:
            data[self._serialize_mapping.get(self._name_attr, self._name_attr)] = (
                getattr(self, self._name_attr)
            )
        for attr in self._get_by_name_attrs:
            name = getattr(self, attr, None)
            if not name:  # pragma: no cover
                raise InitializationError(
                    f"Missing required keyword arg, {attr!r}, to fetch object by name"
                )
            data[self._serialize_mapping.get(attr, attr)] = name
        self.__dict__ = self._credmgr.post(
            f"{self._path}/{self._get_by_name_path}", data=data
        ).__dict__

    def _fetch(self, by_name=False):
        if by_name and self._can_fetch_by_name:
            self._get_by_name()
        else:
            self._get(self.id)
        self._fetched = True

    @_resolve_user
    @warn_camel_case("batchSize")
    def list_items(self, batch_size=20, limit=None, owner=None):
        """List items that are owned by ``owner``."""
        from credmgr.models.helpers import Paginator

        return Paginator(
            self._credmgr,
            self.__class__,
            batch_size=batch_size,
            limit=limit,
            items_owner=owner,
        )

    def __repr__(self):  # pragma: no cover
        """Return repr(self)."""
        return f"{self.__class__.__name__}(id={self.id}, {self._name_attr}={getattr(self, self._name_attr)!r})"

    def __eq__(self, other):
        """Check if both objects are equal."""
        if not isinstance(other, type(self)):
            return False

        return self.id == other.id and getattr(self, self._name_attr) == getattr(
            other, other._name_attr
        )



import pytest

from credmgr.exceptions import InitializationError


def test_user_helper_get_by_id(recorder, credential_manager):
    user = credential_manager.user(1)
    assert getattr(user, user._name_attr) == "spaz"


def test_user_helper_get_by_name(recorder, credential_manager):
    user = credential_manager.user("spaz")
    assert getattr(user, user._name_attr) == "spaz"


def test_user_verification_helper_get_by_id(recorder, credential_manager):
    user_verification = credential_manager.user_verification("0123456789")
    assert getattr(user_verification, user_verification._name_attr) == "0123456789"


def test_user_verification_helper_get_by_name(recorder, credential_manager):
    data = {"user_id": "123456", "reddit_app_id": 22}
    user_verification = credential_manager.user_verification(**data)
    for key, value in data.items():
        assert getattr(user_verification, key) == value


def test_refresh_token_helper_get_by_id(recorder, credential_manager):
    refresh_token = credential_manager.refresh_token(
        redditor="Lil_SpazJoekp", reddit_app_id=2
    )
    assert getattr(refresh_token, refresh_token._name_attr) == "Lil_SpazJoekp"


def test_refresh_token_helper_get_by_name(recorder, credential_manager):
    data = {"redditor": "Lil_SpazJoekp", "reddit_app_id": 2}
    refresh_token = credential_manager.refresh_token(**data)
    for key, value in data.items():
        assert getattr(refresh_token, key) == value


def test_refresh_token_helper_get_by_name_fail(recorder, credential_manager):
    with pytest.raises(InitializationError):
        _ = credential_manager.refresh_token("Lil_SpazJoekp")


def test_refresh_token_helper_get_by_name_missing_name(recorder, credential_manager):
    data = {"reddit_app_id": 2}
    with pytest.raises(InitializationError):
        _ = credential_manager.refresh_token(**data)


def test_sentry_token_helper_get_by_id(recorder, credential_manager):
    sentry_token = credential_manager.sentry_token(1)
    assert getattr(sentry_token, sentry_token._name_attr) == "test"


def test_reddit_app_helper_get_by_id(recorder, credential_manager):
    reddit_app = credential_manager.reddit_app(1)
    assert getattr(reddit_app, reddit_app._name_attr) == "DankMemesModAbuse"


def test_database_credential_helper_get_by_id(recorder, credential_manager):
    database_credential = credential_manager.database_credential(1)
    assert getattr(database_credential, database_credential._name_attr) == "test"


def test_bot_helper_get_by_id(recorder, credential_manager):
    bot = credential_manager.bot(1)
    assert getattr(bot, bot._name_attr) == "test"


def test_bot_helper_get_by_name(recorder, credential_manager):
    bot = credential_manager.bot("test")
    assert getattr(bot, bot._name_attr) == "test"


def test_base_helper_call_with_name_fail(recorder, credential_manager):
    with pytest.raises(InitializationError):
        _ = credential_manager.reddit_app("test")



import configparser

import pytest
from betamax import Betamax

from credmgr import CredentialManager, config
from credmgr.exceptions import InitializationError
from tests.utils import gen_cassette_name


def _initialize_config_override(self):
    self.server = self._fetch("server")
    self.endpoint = self._fetch("endpoint")
    self.api_token = None
    self.username = self._fetch("username")
    self.password = self._fetch("password")


def test_credmgr_init(recorder, credential_manager):
    assert credential_manager.current_user.username == "spaz"


def test_credmgr_init_bad_params(credential_manager):
    with pytest.raises(InitializationError):
        _ = CredentialManager(
            api_token="token", username="username", password="password"
        )


def test_credmgr_init_bad_section_name(credential_manager):
    with pytest.raises(configparser.NoSectionError):
        _ = CredentialManager("invalid")


def test_credmgr_no_params():
    old = config.Config._initialize_config
    config.Config._initialize_config = _initialize_config_override
    with pytest.raises(InitializationError):
        _ = CredentialManager()
    config.Config._initialize_config = old


def test_credmgr_camel_case():
    credmgr = CredentialManager(api_token="token")
    assert credmgr.config.api_token == "token"

    with pytest.deprecated_call():
        credmgr = CredentialManager(api_token="token", apiToken="old token")
        assert credmgr.config.api_token == "token"


def test_credmgr_init_username_password():
    old = config.Config._initialize_config
    config.Config._initialize_config = _initialize_config_override
    credmgr = CredentialManager(username="username", password="password")
    with Betamax(credmgr._requestor._session).use_cassette(gen_cassette_name()):
        assert credmgr.current_user.username == "username"
    config.Config._initialize_config = old



"""Provide the credmgr models."""

from .bot import Bot  # NOQA
from .database_credential import DatabaseCredential  # NOQA
from .reddit_app import RedditApp  # NOQA
from .refresh_token import RefreshToken  # NOQA
from .sentry_token import SentryToken  # NOQA
from .user import User  # NOQA
from .user_verification import UserVerification  # NOQA

from .helpers import (  # NOQA isort: skip
    BotHelper,
    DatabaseCredentialHelper,
    RedditAppHelper,
    RefreshTokenHelper,
    SentryTokenHelper,
    UserHelper,
    UserVerificationHelper,
)



"""Provide the DatabaseCredential class."""

from ..deprecation_decorator import warn_camel_case
from ..mixins import BaseApp
from .utils import _resolve_user


class DatabaseCredential(BaseApp):
    """A class for DatabaseCredential instances."""

    _credmgr_callable = "database_credential"
    _editable_attrs = BaseApp._editable_attrs + [
        "database",
        "database_flavor",
        "database_host",
        "database_password",
        "database_port",
        "database_username",
        "private_key",
        "private_key_passphrase",
        "ssh_host",
        "ssh_password",
        "ssh_port",
        "ssh_username",
        "use_ssh",
        "use_ssh_key",
    ]
    _path = "/database_credentials"

    @classmethod
    def schema(cls):
        """Return the schema for this model."""
        if cls._schema is None:
            from .schemas import DatabaseCredentialSchema

            cls._schema = DatabaseCredentialSchema()
        return cls._schema

    def __init__(self, credmgr, **kwargs):
        """Initialize a Database Credential instance.

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param id: The ID of the Database Credential.
        :param str name: Name of the Database Credential.
        :param str database_flavor: Type of database.
        :param str database: Working database to use.
        :param str database_host: Database server address.
        :param int database_port: Port the database server listens on.
        :param str database_username: Username used to connect and authenticate the
            database.
        :param str database_password: Password used to connect and authenticate the
            database.
        :param bool use_ssh: Determines if the database will be connected to through a
            tunnel.
        :param str ssh_host: The address of the server that the SSH tunnel will connect
            to.
        :param str ssh_port: The port the SSH tunnel will use.
        :param str ssh_username: Username for the SSH tunnel.
        :param str ssh_password: Password for the SSH tunnel.
        :param bool use_ssh_key: Determines if the SSH tunnel will use private key
            authentication.
        :param str private_key: SSH private key. Note: No validation will be performed.
        :param str private_key_passphrase: Passphrase for the SSH key.
        :param owner_id: ID of the `.User` that owns this Database Credential.

        """
        super().__init__(credmgr, **kwargs)

    @staticmethod
    @_resolve_user
    @warn_camel_case(
        "databaseFlavor",
        "database",
        "databaseHost",
        "databasePort",
        "databaseUsername",
        "databasePassword",
        "useSsh",
        "sshHost",
        "sshPort",
        "sshUsername",
        "sshPassword",
        "useSshKey",
        "privateKey",
        "privateKeyPassphrase",
    )
    def _create(
        _credmgr,
        name,
        database_flavor="postgres",
        database="postgres",
        database_host="localhost",
        database_port=5432,
        database_username="postgres",
        database_password=None,
        use_ssh=False,
        ssh_host=None,
        ssh_port=None,
        ssh_username=None,
        ssh_password=None,
        use_ssh_key=False,
        private_key=None,
        private_key_passphrase=None,
        enabled=True,
        owner=None,
    ):
        """Create a new Database Credential.

        Database Credentials are used for..ya know..databases

        :param str name: Name of the Database Credential (required)
        :param str database_flavor: Type of database, (default: ``postgres``)
        :param str database: Working database to use, (default: ``postgres``)
        :param str database_host: Database server address, (default: ``localhost``)
        :param int database_port: Port the database server listens on, (default:
            ``5432``)
        :param str database_username: Username to use to connect to the database
        :param str database_password: Password to use to connect to the database
        :param bool use_ssh: Determines if the database will be connected to through a
            tunnel
        :param str ssh_host: The address of the server that the SSH tunnel will connect
            to
        :param str ssh_port: The port the SSH tunnel will use
        :param str ssh_username: Username for the SSH tunnel
        :param str ssh_password: Password for the SSH tunnel
        :param bool use_ssh_key: Allows the credentials to be used
        :param str private_key: SSH private key. Note: No validation will be performed.
        :param str private_key_passphrase: Passphrase for the SSH key
        :param bool enabled: Allows the credentials to be used
        :param Union[User,int,str] owner: Owner of the Database Credential. Requires
            Admin to create for other users.

        :returns: DatabaseCredential

        """
        data = {}
        if database_flavor:
            data["database_flavor"] = database_flavor
        if database:
            data["database"] = database
        if database_host:
            data["database_host"] = database_host
        if database_port:
            data["database_port"] = database_port
        if database_username:
            data["database_username"] = database_username
        if database_password:
            data["database_password"] = database_password
        if use_ssh:
            data["use_ssh"] = use_ssh
        if ssh_host:
            data["ssh_host"] = ssh_host
        if ssh_port:
            data["ssh_port"] = ssh_port
        if ssh_username:
            data["ssh_username"] = ssh_username
        if ssh_password:
            data["ssh_password"] = ssh_password
        if use_ssh_key:
            data["use_ssh_key"] = use_ssh_key
        if private_key:
            data["private_key"] = private_key
        if private_key_passphrase:
            data["private_key_passphrase"] = private_key_passphrase
        if enabled:
            data["enabled"] = enabled
        if owner:
            data["owner_id"] = owner
        return _credmgr.post("/database_credentials", data={"app_name": name, **data})



"""Provide the exception for credmgr."""


class CredentialManagerException(Exception):
    """The base CredentialManager Exception that all other exception classes extend."""


class InitializationError(CredentialManagerException):
    """Invalid initialization parameters were used."""


class SerializerException(CredentialManagerException):
    """Serialization/deserialization error occurred."""


class APIException(CredentialManagerException):
    """Base exception class."""

    def __init__(self, response):
        """Initialize an APIException instance."""
        self.url = response.url
        self.status = response.status_code
        self.reason = response.reason
        self.body = response.text
        self.headers = response.headers

    def __str__(self):  # pragma: no cover
        """Get the message returned from str(self)."""
        error_message = f"\n{self.status}: {self.url}\n{self.reason}"
        if self.headers:
            error_message += f"\n_response headers: {self.headers}"
        if self.body:
            error_message += f"\n_response body: {self.body}"

        return error_message


class Conflict(APIException):
    """Made a conflicting request."""


class Forbidden(APIException):
    """Made request to a resource without required permissions."""


class InvalidJSON(APIException):
    """Response returned invalid JSON."""


class InvalidRequest(APIException):
    """Made request with invalid parameters."""


class NotFound(APIException):
    """Could not find requested resource."""


class ServerError(APIException):
    """Error occurred on server during request."""


class Unauthorized(APIException):
    """Made request to resource without authenticating."""


class UnknownStatusCode(APIException):
    """Received an unknown status code."""



import pytest

from credmgr.exceptions import NotFound
from credmgr.models import RefreshToken


def test_delete_refresh_token(recorder, credential_manager):
    refresh_token = credential_manager.refresh_token(id=1)
    refresh_token.delete()
    with pytest.raises(NotFound):
        _ = credential_manager.refresh_token(id=1)


def test_list_refresh_tokens(recorder, credential_manager):
    refresh_tokens = credential_manager.refresh_tokens()
    for refresh_token in refresh_tokens:
        assert isinstance(refresh_token, RefreshToken)



"""Provide the Requestor class."""

import logging
from collections import defaultdict

import requests
from requests import codes

from .const import __version__
from .exceptions import (
    Conflict,
    Forbidden,
    InvalidJSON,
    InvalidRequest,
    NotFound,
    ServerError,
    Unauthorized,
    UnknownStatusCode,
)

log = logging.getLogger(__package__)


def _urljoin(base, path):
    if base.endswith("/"):  # pragma: no cover
        base = base[:-1]
    if not path.startswith("/"):  # pragma: no cover
        path = f"/{path}"
    return f"{base}{path}"


class Requestor:
    """Requestor provides an interface to HTTP requests."""

    _exception_mapping = defaultdict(
        lambda: UnknownStatusCode,
        {
            400: InvalidRequest,
            401: Unauthorized,
            403: Forbidden,
            404: NotFound,
            409: Conflict,
            422: ServerError,
            500: ServerError,
            502: ServerError,
            503: ServerError,
            504: ServerError,
            520: ServerError,
            522: ServerError,
        },
    )
    _retry_error = (InvalidRequest, ServerError, UnknownStatusCode)
    _success_codes = [200, 201, 202, 204]

    def __init__(self, credmgr_url, auth, session_class, **session_kwargs):
        """Create an instance of the Requestor class.

        :param str credmgr_url: Url used to make API requests
        :param auth: An auth tuple or a class that subclasses requests.auth
        :param Session session_class: (Optional) A custom session class to handle
            requests, compatible with requests.Session(). (Default: None)

        """
        self._base_url = credmgr_url
        self._session = (session_class or requests.Session)()
        self._session.auth = auth
        for key, value in session_kwargs.items():  # pragma: no cover
            setattr(self._session, key, value)
        self._session.headers["User-Agent"] = f"credmgr/{__version__}"

    def __getattr__(self, attribute):  # pragma: no cover
        """Pass all undefined attributes to the _session attribute."""
        if attribute.startswith("__"):
            raise AttributeError
        return getattr(self._session, attribute)

    @staticmethod
    def _log_request(data, method, params, url):
        log.debug(f"Data: {data}")
        log.debug(f"Request: {method} {url}")
        log.debug(f"Query Parameters: {params}")

    def request(self, path, method, data=None, params=None, **kwargs):
        """Issue the HTTP request capturing any errors that may occur."""
        url = _urljoin(self._base_url, path)
        retry_limit = 4
        retry_count = 0
        while retry_count < retry_limit:
            try:
                self._log_request(data, method, params, url)
                response = self._session.request(
                    method, url, params, data=data, timeout=15, **kwargs
                )
                log.debug(
                    f"Response: {response.status_code} ({response.headers.get('content-length')} bytes)"
                )
                if response.status_code in self._exception_mapping:
                    raise self._exception_mapping[response.status_code](response)
                elif response.status_code == codes["no_content"]:
                    return
                if response.headers.get("content-length") == "0":  # pragma: no cover
                    return ""
                try:
                    response.json()
                except ValueError:  # pragma: no cover
                    raise InvalidJSON(response) from None
                return response
            except self._retry_error as error:
                retry_count += 1
                if retry_count < retry_limit:
                    log.debug(
                        f"Error occurred: {error}. Retrying...attempt {retry_count}/{retry_limit - 1}"
                    )
                else:
                    log.error(f"Error occurred: {error}. Max attempts reached.")
                    raise error



"""Provide the User class."""

import json
from datetime import datetime

from ..deprecation_decorator import warn_camel_case
from ..exceptions import InitializationError
from ..mixins import BaseModel, DeletableMixin, EditableMixin


class User(DeletableMixin, EditableMixin):
    """A class for User instances."""

    _attr_types = {
        **BaseModel._attr_types,
        "created": datetime,
        "updated": datetime,
    }
    _can_fetch_by_name = True
    _credmgr_callable = "user"
    _editable_attrs = [
        "default_settings",
        "is_active",
        "is_admin",
        "is_regular_user",
        "reddit_username",
        "username",
    ]
    _name_attr = "username"
    _path = "/users"

    @classmethod
    def schema(cls):
        """Return the schema for this model."""
        if cls._schema is None:
            from .schemas import UserSchema

            cls._schema = UserSchema()
        return cls._schema

    def __init__(self, credmgr, **kwargs):
        """Initialize an User instance.

        Users are for logging into and accessing that user's credentials

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param int id: ID of the User.
        :param username: Username of the User.
        :param bool is_active: Indicates if the User can log in and access
            CredentialManager.
        :param bool is_regular_user: Indicates if this is a regular user.
        :param bool is_admin: Indicates if this User can create other users and their
            credentials.
        :param default_settings: Default settings used when creating new items without
            those settings explicitly set.
        :param str reddit_username: This User's Reddit username. Used for
            :class:`.RedditApp`'s user_agent.
        :param datetime.datetime created: Date and time this User was created.
        :param datetime.datetime updated: Date and time this User was last updated.
        :param list[RedditApp] reddit_apps: A list of Reddit Apps this User owns.
        :param list[SentryToken] sentry_tokens: A list of Sentry Tokens this User owns.
        :param list[DatabaseCredential] database_credentials: A list of Database
            Credentials this User owns.

        """
        super().__init__(credmgr, **kwargs)
        self._apps = {}
        if "reddit_apps" in kwargs:
            self._apps["reddit_apps"] = kwargs["reddit_apps"]
        if "sentry_tokens" in kwargs:
            self._apps["sentry_tokens"] = kwargs["sentry_tokens"]
        if "database_credentials" in kwargs:
            self._apps["database_credentials"] = kwargs["database_credentials"]

    @staticmethod
    @warn_camel_case(
        "defaultSettings",
        "redditUsername",
        "isAdmin",
        "isActive",
        "isRegularUser",
        "isInternal",
    )
    def _create(
        _credmgr,
        username,
        password,
        default_settings=None,
        reddit_username=None,
        is_admin=False,
        is_active=True,
        is_regular_user=True,
        is_internal=False,
    ):
        """Create a new User.

        **PERMISSIONS: Admin role is required.**

        :param str username: Username for new user (Example: ``spaz``) (required)
        :param str password: Password for new user (Example: ``supersecurepassword``)
            (required)
        :param dict default_settings: Default values to use for new apps (Example:
            ``{"database_flavor": "postgres", "database_host": "localhost"}``)
        :param str reddit_username: User's Reddit username (Example: ``Lil_SpazJoekp``)
        :param bool is_admin: Is the user an admin? Allows the user to see all objects
            and create users (Default: ``False``)
        :param bool is_active: Is the user active? Allows the user to sign in (Default:
            ``True``)
        :param bool is_regular_user: (Internal use only)
        :param bool is_internal: (Internal use only)

        :returns: User

        """
        additional_params = {}
        if default_settings:
            additional_params["default_settings"] = json.dumps(default_settings)
        if is_admin:  # pragma: no cover
            additional_params["is_admin"] = is_admin
        if is_active:
            additional_params["is_active"] = is_active
        if is_regular_user:
            additional_params["is_regular_user"] = is_regular_user
        if is_internal:  # pragma: no cover
            additional_params["is_internal"] = is_internal
        if reddit_username:
            additional_params["reddit_username"] = reddit_username
        return _credmgr.post(
            "/users",
            data={"username": username, "password": password, **additional_params},
        )

    def apps(self, only=None):
        """Get apps that are owned by this user.

        :param str only: Pass one of ``reddit_apps``, ``sentry_tokens``,
            ``database_credentials`` to only get that type of apps

        :returns: The user's apps.

        """
        app_types = ["reddit_apps", "sentry_tokens", "database_credentials"]
        if only and only not in app_types:
            raise InitializationError(
                f"App type: {only} is not valid. Only 'reddit_apps', 'sentry_tokens', and 'database_credentials' are valid."
            )
        if not self._apps:
            response = self._credmgr.get(f"/users/{self.id}/apps")
            self._apps = response._apps
            for app in app_types:
                setattr(self, app, self._apps[app])
        return self._apps[only] if only else self._apps



import pytest

from credmgr.exceptions import Conflict, NotFound
from credmgr.models import UserVerification


def test_create_user_verification(recorder, credential_manager):
    data = {"user_id": "12345", "reddit_app": 27}
    user_verification = credential_manager.user_verification.create(**data)
    for key, value in data.items():
        if key == "reddit_app":
            assert getattr(user_verification, key).id == value
            continue
        assert getattr(user_verification, key) == value


def test_create_user_verification_full_params(recorder, credential_manager):
    data = {
        "user_id": "1234567",
        "reddit_app": 27,
        "redditor": "redditor",
        "extra_data": {"extra": "data", "nested": {"dict": "value"}},
    }
    user_verification = credential_manager.user_verification.create(**data)
    for key, value in data.items():
        if key == "reddit_app":
            assert getattr(user_verification, key).id == value
            continue
        assert getattr(user_verification, key) == value


def test_create_user_verification_other_user(recorder, credential_manager):
    data = {"user_id": "123456", "reddit_app": 22, "owner": 4}
    user_verification = credential_manager.user_verification.create(**data)
    for key, value in data.items():
        if key in ["owner", "reddit_app"]:
            assert getattr(user_verification, key).id == value
            continue
        assert getattr(user_verification, key) == value


def test_create_user_verification_existing(recorder, credential_manager):
    data = {"user_id": "12345", "reddit_app": 27}
    existing = credential_manager.user_verification.create(**data)
    assert isinstance(existing, UserVerification)
    assert existing.user_id == "12345"


def test_delete_user_verification(recorder, credential_manager):
    user_verification = credential_manager.user_verification("12345")
    user_verification.delete()
    with pytest.raises(NotFound):
        _ = credential_manager.user_verification("12345")


def test_edit_user_verification(recorder, credential_manager):
    user_verification = credential_manager.user_verification("new_id")
    user_verification.edit(redditor="redditor")
    assert user_verification.redditor == "redditor"


def test_edit_user_verification_conflicting_data(recorder, credential_manager):
    user_verification = credential_manager.user_verification("new_id")
    with pytest.raises(Conflict):
        user_verification.edit(user_id="12345")


def test_list_user_verifications(recorder, credential_manager):
    user_verifications = credential_manager.user_verifications()
    for user_verification in user_verifications:
        assert isinstance(user_verification, UserVerification)



"""Provide helper utilities used by other models."""

from functools import wraps

from credmgr.exceptions import NotFound


def _resolve_model_from_input(credmgr, model, input_value, return_attr="id"):
    value = None
    if isinstance(input_value, model):
        value = getattr(input_value, return_attr)
    elif isinstance(input_value, int):
        value = input_value
    elif isinstance(input_value, str):
        try:
            found_item = getattr(credmgr, model._credmgr_callable)(input_value)
        except NotFound:
            found_item = None
        if found_item:
            value = getattr(found_item, return_attr)
    return value


def _resolve_user(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        from . import User

        value = None
        user = kwargs.get("owner", None)
        if user:
            if isinstance(user, User):
                value = user.id
            elif isinstance(user, int):
                value = user
            elif isinstance(user, str):
                found_user = self._credmgr.user(user)
                if found_user:
                    value = found_user.id
            kwargs["owner"] = value
        return func(self, *args, **kwargs)

    return wrapper


try:
    from functools import cached_property
except ImportError:  # pragma: no cover
    from cached_property import cached_property  # noqa: F401



import pytest

from credmgr.exceptions import Conflict, NotFound
from credmgr.models import Bot


def assert_created(bot, data):
    for key, value in data.items():
        if key in ["reddit_app", "sentry_token", "database_credential", "owner"]:
            assert getattr(bot, key).id == value
            continue
        assert getattr(bot, key) == value


def test_create_bot(recorder, credential_manager):
    data = {"name": "bot_name"}
    bot = credential_manager.bot.create(**data)
    assert_created(bot, data)


def test_create_bot_full_params(recorder, credential_manager):
    data = {
        "name": "bot_name_2",
        "reddit_app": 27,
        "sentry_token": 4,
        "database_credential": 1,
    }
    bot = credential_manager.bot.create(**data)
    assert_created(bot, data)


def test_create_bot_full_params_objects(recorder, credential_manager):
    data = {
        "name": "bot_name_2",
        "reddit_app": credential_manager.reddit_app(27),
        "sentry_token": credential_manager.sentry_token(4),
        "database_credential": credential_manager.database_credential(1),
    }
    bot = credential_manager.bot.create(**data)
    for key, value in data.items():
        assert getattr(bot, key) == value


def test_create_bot_other_user(recorder, credential_manager):
    data = {
        "name": "bot_name_3",
        "reddit_app": 29,
        "sentry_token": 12,
        "database_credential": 20,
        "owner": 4,
    }
    bot = credential_manager.bot.create(**data)
    assert_created(bot, data)


def test_create_bot_existing(recorder, credential_manager):
    data = {"name": "bot_name", "reddit_app": 27}
    with pytest.raises(Conflict):
        _ = credential_manager.bot.create(**data)


def test_delete_bot(recorder, credential_manager):
    bot = credential_manager.bot("bot_name")
    bot.delete()
    with pytest.raises(NotFound):
        _ = credential_manager.bot("bot_name")


def test_edit_bot(recorder, credential_manager):
    bot = credential_manager.bot("bot_name_2")
    bot.edit(reddit_app=28, name="new_bot_name")
    assert bot.reddit_app.id == 28
    assert bot.name == "new_bot_name"


def test_edit_bot_with_object(recorder, credential_manager):
    bot = credential_manager.bot("bot_name4")
    bot.edit(reddit_app=credential_manager.reddit_app(28))
    assert bot.reddit_app.id == 28


def test_edit_bot_conflicting_data(recorder, credential_manager):
    bot = credential_manager.bot("new_bot_name")
    with pytest.raises(Conflict):
        bot.edit(name="bot_name4")


def test_list_bots(recorder, credential_manager):
    bots = credential_manager.bots()
    for bot in bots:
        assert isinstance(bot, Bot)



from credmgr import CredentialManager

credential_manager = CredentialManager()

bot = credential_manager.bot("bot_name")
reddit = bot.reddit_app.reddit("authorized_redditor")

subreddit = reddit.subreddit("dankmemes")
stream = subreddit.stream.comments()

while True:
    for comment in stream:
        if "mods gay" in comment.body and not comment.saved:
            comment.reply("no u")
            comment.save()



"""Provide the SentryToken class."""

from ..mixins import BaseApp
from .utils import _resolve_user


class SentryToken(BaseApp):
    """A class for SentryToken instances."""

    _credmgr_callable = "sentry_token"
    _editable_attrs = BaseApp._editable_attrs + ["dsn"]
    _path = "/sentry_tokens"

    @classmethod
    def schema(cls):
        """Return the schema for this model."""
        if cls._schema is None:
            from .schemas import SentryTokenSchema

            cls._schema = SentryTokenSchema()
        return cls._schema

    def __init__(self, credmgr, **kwargs):
        """Initialize a SentryToken instance.

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param id: ID of the Sentry Token
        :param str name: Name of the Sentry Token.
        :param str dsn: Sentry DSN url for reporting errors.
        :param owner_id: ID of the `.User` that owns this Sentry Token.

        """
        super().__init__(credmgr, **kwargs)

    @staticmethod
    @_resolve_user
    def _create(_credmgr, name=None, dsn=None, owner=None):
        """Create a new Sentry Token.

        Sentry Tokens are used for logging and error reporting in applications

        :param str name: Name of the Sentry Token (required)
        :param str dsn: DSN of the Sentry Token (required)
        :param Union[User,int,str] owner: Owner of the verification. Requires Admin to
            create for other users.

        :returns: SentryToken

        """
        data = {"app_name": name, "dsn": dsn}
        if owner:
            data["owner_id"] = owner
        return _credmgr.post("/sentry_tokens", data=data)



"""Provide the Bot class."""

from ..deprecation_decorator import warn_camel_case
from ..mixins import BaseApp
from .utils import _resolve_user


class Bot(BaseApp):
    """A class for Bot instances.

    To obtain an instance of this class execute:

    .. code-block:: python

        bot = credmgr.bot("bot_name")

    Bots is the intended way to use CredentialManager. Bots are for grouping credentials
    that would be used in a single app.

    The following is the intended usage:

    .. code-block:: python

        from credmgr import CredentialManager

        credential_manager = CredentialManager(api_token="api_token")
        bot = credmgr.bot("bot_name")

        reddit = bot.reddit("Lil_SpazJoekp")
        db_creds = bot.database_credentials
        sentryDSN = bot.sentry_token

    """

    _can_fetch_by_name = True
    _credmgr_callable = "bot"
    _editable_attrs = BaseApp._editable_attrs + [
        "database_credential_id",
        "reddit_app_id",
        "sentry_token_id",
    ]
    _path = "/bots"

    @classmethod
    def schema(cls):
        """Return the schema for this model."""
        if cls._schema is None:
            from .schemas import BotSchema

            cls._schema = BotSchema()
        return cls._schema

    def __init__(self, credmgr, **kwargs):
        """Initialize a Bot instance.

        Bots are used for grouping credentials into a single app

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param id: ID of this Bot.
        :param name: Name of this Bot.
        :param owner_id: ID of the `.User` that owns this Bot.
        :param reddit_app: `.RedditApp` that will be used with this Bot.
        :param sentry_token: `.SentryToken` that will be used with this Bot.
        :param database_credential: `.DatabaseCredential` that will be used with this
            Bot.

        """
        super().__init__(credmgr, **kwargs)

    @staticmethod
    @_resolve_user
    @warn_camel_case("redditApp", "sentryToken", "databaseCredential")
    def _create(
        _credmgr, name, reddit_app, sentry_token, database_credential, owner=None
    ):
        """Create a new Bot.

        Bots are used for grouping credentials into a single app

        :param str name: Name of the Bot (required)
        :param Union[RedditApp,int] reddit_app: Reddit App the bot will use
        :param Union[SentryToken,int] sentry_token: Sentry Token the bot will use
        :param Union[DatabaseCredential,int] database_credential: Database Credentials
            the bot will use
        :param Union[User,int,str] owner: Owner of the bot. Requires Admin to create for
            other users.

        :returns: Bot

        """
        from . import DatabaseCredential, RedditApp, SentryToken

        additional_params = {}
        existing_objects = {}
        if isinstance(reddit_app, RedditApp):
            existing_objects["reddit_app"] = reddit_app
            reddit_app = reddit_app.id
        if reddit_app:
            additional_params["reddit_app_id"] = reddit_app
        if isinstance(sentry_token, SentryToken):
            existing_objects["sentry_token"] = sentry_token
            sentry_token = sentry_token.id
        if sentry_token:
            additional_params["sentry_token_id"] = sentry_token
        if isinstance(database_credential, DatabaseCredential):
            existing_objects["database_credential"] = database_credential
            database_credential = database_credential.id
        if database_credential:
            additional_params["database_credential_id"] = database_credential
        if owner:
            additional_params["owner_id"] = owner
        response = _credmgr.post("/bots", data={"app_name": name, **additional_params})
        for key, value in existing_objects.items():
            setattr(response, key, value)
        return response

    def edit(self, **kwargs):
        """Edit the bot.

        :param name: Changes the name of the :class:`.Bot`
        :param Union[RedditApp,int] reddit_app: Changes the :class:`.RedditApp` the Bot
            will use
        :param Union[SentryToken,int] sentry_token: Changes the :class:`.SentryToken`
            the Bot will use
        :param Union[DatabaseCredential,int] database_credential: Changes the
            :class:`.DatabaseCredential` the Bot will use

        .. note::

            Parameters, ``reddit_app``, ``sentry_token``, and ``database_credential``
            can accept the initialized object or its :attr:`id`. Passing a ``str`` to it
            will not work.

        :returns: The modified :class:`.Bot`

        """
        from . import DatabaseCredential, RedditApp, SentryToken

        for key, _value in list(kwargs.items()):
            if key in ["reddit_app", "sentry_token", "database_credential"]:
                if isinstance(
                    kwargs[key], (RedditApp, SentryToken, DatabaseCredential)
                ):
                    new_key = f"{key}_id"
                    kwargs[new_key] = kwargs.pop(key).id
                else:
                    new_key = f"{key}_id"
                    kwargs[new_key] = kwargs.pop(key)
        super().edit(**kwargs)



"""Provide the OwnerMixin class."""

from credmgr.mixins import BaseModel
from credmgr.models.utils import cached_property


class OwnerMixin(BaseModel):
    """Interface for classes that have an owner."""

    _editable_attrs = []

    @cached_property
    def owner(self):
        """Get the owner of the object."""
        if "owner_id" not in self.__dict__:
            if not self._fetched:
                self._fetch()
        user = self._credmgr.user(id=self.owner_id)
        return user



"""Provide the helper classes."""

from ..deprecation_decorator import warn_camel_case
from ..exceptions import InitializationError
from ..mixins import BaseModel
from . import (
    Bot,
    DatabaseCredential,
    RedditApp,
    RefreshToken,
    SentryToken,
    User,
    UserVerification,
)
from .utils import _resolve_user


class Paginator:
    """Class to handle paginated requests."""

    @_resolve_user
    def __init__(self, credmgr, model, batch_size=20, limit=None, items_owner=None):
        """Initialize a Paginator instance.

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param model: A CredentialManager model to list.
        :param batch_size: The number of items to fetch at a time. If ``batch_size`` is
            None, it will fetch them 100 at a time. (default: 20).
        :param limit: The maximum number of items to get.
        :param items_owner: Owner to filter the items by.

        """
        self._completed = False
        self._credmgr = credmgr
        self._current_item_index = None
        self._items_returned = 0
        self._model = model(self._credmgr)
        self._offset = 0
        self._response = None
        self.batch_size = batch_size
        self.items_owner = items_owner
        self.limit = limit

    def __iter__(self):
        """Allow Paginator to operate as an iterator."""
        return self

    def __next__(self):
        """Allow Paginator to operate as an generator."""
        if self.limit is not None and self._items_returned >= self.limit:
            raise StopIteration()  # pragma: no cover

        if self._response is None or self._current_item_index >= len(self._response):
            self._next()

        self._current_item_index += 1
        self._items_returned += 1
        return self._response[self._current_item_index - 1]

    def _next(self):
        if self._completed:
            raise StopIteration()
        params = {"limit": self.batch_size, "offset": self._offset}
        if self.items_owner:
            params["owner_id"] = self.items_owner
        self._response = self._credmgr.get(self._model._path, params=params)
        self._current_item_index = 0
        if not self._response:
            raise StopIteration()  # pragma: no cover
        if self._response and len(self._response) == self.batch_size:
            self._offset += self.batch_size  # pragma: no cover
        else:
            self._completed = True


class BaseHelper(BaseModel):
    """Base Helper class."""

    _model = None

    @warn_camel_case()
    def __call__(self, id=None, **kwargs):
        """Fetch and return an instance of the object."""
        model_kwargs = {}
        by_name = False
        if isinstance(id, str):
            name = id
            id = None
            if self._model._can_fetch_by_name:
                by_name = True
                kwargs[self._model._name_attr] = name
            else:
                raise InitializationError(
                    f"Cannot get {self._model.__name__!r} by name"
                )
        if id:
            model_kwargs["id"] = id
        elif self._model._can_fetch_by_name:
            by_name = True
            if not self._model._get_by_name_attrs:
                model_kwargs[self._model._name_attr] = kwargs[self._model._name_attr]
            for key in self._model._get_by_name_attrs:
                name = kwargs.get(key, None)
                if name:
                    model_kwargs[key] = name
                else:
                    raise InitializationError(
                        f"Missing required keyword argument, {key!r}, to fetch object by name"
                    )
        else:
            raise InitializationError("'id' is required")
        item = self._model(self._credmgr, **model_kwargs)
        item._fetch(by_name)
        return item


class UserHelper(BaseHelper):
    """Interface for interacting with Users."""

    _model = User

    def __call__(self, id=None, username=None):
        """Fetch a :class:`.User` instance by :attr:`id` or :attr:`name`.

        :param Union[int,str] id: ID of the :class:`.User` to fetch. .. note:: If a
            ``str`` is passed it will be treated as the :attr:`username` attr.
        :param str username: Username of the :class:`.User` to fetch. .. note:: If both
            :attr:`id` and :attr:`username` are passed the :attr:`id` will take
            precedence. If :attr:`id` is a ``str`` it will be treated as an
            :attr:`username` and will also take precedence.

        :returns: An initialized :class:`.User` instance.

        """
        kwargs = {}
        by_name = False
        if isinstance(id, str):
            by_name = True
            username = id
            id = None
        if id:
            kwargs["id"] = id
        if username:
            kwargs["username"] = username
        if not (id or username):
            raise InitializationError("At least 'id' or 'username' is required")
        item = User(self._credmgr, **kwargs)
        item._fetch(by_name)
        return item

    def create(
        self,
        username,
        password,
        default_settings=None,
        is_admin=False,
        is_active=True,
        is_regular_user=True,
        is_internal=False,
        reddit_username=None,
    ) -> User:
        """Create a new user.

        **PERMISSIONS: At least ``is_admin`` is required.**

        :param str username: Username for new user (Example: ``spaz``) (required)
        :param str password: Password for new user (Example: ``supersecurepassword``)
            (required)
        :param dict default_settings: Default values to use for new apps (Example:
            ``{"database_flavor": "postgres", "database_host": "localhost"}``)
        :param bool is_admin: Is the user an admin? Allows the user to see all objects
            and create users (Default: ``False``)
        :param bool is_active: Is the user active? Allows the user to sign in (Default:
            ``True``)
        :param bool is_regular_user: (Internal use only)
        :param bool is_internal: (Internal use only)
        :param str reddit_username: The user's Reddit username.

        :returns: User

        """
        return self._model._create(
            self._credmgr,
            username=username,
            password=password,
            default_settings=default_settings,
            is_admin=is_admin,
            is_active=is_active,
            is_regular_user=is_regular_user,
            is_internal=is_internal,
            reddit_username=reddit_username,
        )


class BotHelper(BaseHelper):
    """Interface for interacting with Bots."""

    _model = Bot

    def create(
        self,
        name,
        reddit_app=None,
        sentry_token=None,
        database_credential=None,
        owner=None,
    ) -> Bot:
        """Create a new Bot.

        Bots are used for grouping credentials into a single app.

        :param str name: Name of the Bot (required).
        :param Union[RedditApp,int] reddit_app: Reddit App the bot will use.
        :param Union[SentryToken,int] sentry_token: Sentry Token the bot will use.
        :param Union[DatabaseCredential,int] database_credential: Database Credentials
            the bot will use.
        :param Union[User,int,str] owner: Owner of the bot. Requires Admin to create for
            other users.

        :returns: Bot

        """
        return self._model._create(
            self._credmgr,
            name=name,
            reddit_app=reddit_app,
            sentry_token=sentry_token,
            database_credential=database_credential,
            owner=owner,
        )


class RedditAppHelper(BaseHelper):
    """Interface for interacting with RedditApps."""

    _model = RedditApp

    def create(
        self,
        name,
        client_id,
        user_agent=None,
        app_type="web",
        redirect_uri=None,
        client_secret=None,
        app_description=None,
        enabled=True,
        owner=None,
    ) -> RedditApp:
        """Create a new RedditApp.

        Reddit Apps are used for interacting with reddit.

        :param str name: Name of the RedditApp (required).
        :param str client_id: Client ID of the Reddit App (required).
        :param str user_agent: User agent used for requests to Reddit's API (required,
            defaults to user set default, then to 'python:{name} by
            /u/{reddit_username}' if current_user.reddit_username is set or
            'python:{name}' if it is not set).
        :param str app_type: Type of the app. One of `web`, `installed`, or `script`
            (required, default: 'web').
        :param str redirect_uri: Redirect URI for Oauth2 flow. (required, defaults to
            user set default then to
            `https://credmgr.jesassn.org/oauth2/reddit_callback` if neither are set).
        :param str client_secret: Client secret of the Reddit App.
        :param str app_description: Description of the Reddit App.
        :param bool enabled: Allows the app to be used.
        :param Union[User,int,str] owner: Owner of the Reddit App. Requires Admin to
            create for other users.

        :returns: RedditApp

        """
        if not user_agent:
            reddit_username = self._credmgr.current_user.reddit_username
            reddit_username_str = ""
            if reddit_username:
                reddit_username_str = f" by u/{reddit_username}"
            user_agent = self._credmgr.get_user_default(
                "user_agent", f"python:{name}{reddit_username_str}"
            )
        redirect_uri = self._credmgr.get_user_default(
            "redirect_uri",
            redirect_uri or "https://credmgr.jesassn.org/oauth2/reddit_callback",
        )
        return self._model._create(
            self._credmgr,
            name=name,
            client_id=client_id,
            user_agent=user_agent,
            app_type=app_type,
            redirect_uri=redirect_uri,
            client_secret=client_secret,
            app_description=app_description,
            enabled=enabled,
            owner=owner,
        )


class UserVerificationHelper(BaseHelper):
    """Interface for interacting with UserVerifications."""

    _model = UserVerification

    def create(
        self, user_id, reddit_app, redditor=None, extra_data=None, owner=None
    ) -> UserVerification:
        """Create a new User Verification.

        User verifications for verifying a redditor with a user ID.

        :param str user_id: User ID to associate redditor with (required).
        :param Union[RedditApp,int,str] reddit_app: Reddit app the user verification is
            for (required).
        :param str redditor: Redditor the user verification is for.
        :param dict extra_data: Extra JSON data to include with verification.
        :param int owner: Owner of the verification. Requires Admin to create for other
            users.

        :returns: UserVerification

        """
        return self._model._create(
            self._credmgr,
            user_id=user_id,
            reddit_app=reddit_app,
            redditor=redditor,
            extra_data=extra_data,
            owner=owner,
        )


class SentryTokenHelper(BaseHelper):
    """Interface for interacting with SentryTokens."""

    _model = SentryToken

    def create(self, name, dsn, owner=None) -> SentryToken:
        """Create a new Sentry Token.

        Sentry Tokens are used for logging and error reporting in applications

        :param str name: Name of the Sentry Token (required).
        :param str dsn: DSN of the Sentry Token (required).
        :param Union[User,int,str] owner: Owner of the verification. Requires Admin to
            create for other users.

        :returns: SentryToken

        """
        return self._model._create(self._credmgr, name=name, dsn=dsn, owner=owner)


class DatabaseCredentialHelper(BaseHelper):
    """Interface for interacting with DatabaseCredentials."""

    _model = DatabaseCredential

    def create(
        self,
        name,
        database_flavor="postgres",
        database="postgres",
        database_host="localhost",
        database_port=5432,
        database_username="postgres",
        database_password=None,
        use_ssh=False,
        ssh_host=None,
        ssh_port=None,
        ssh_username=None,
        ssh_password=None,
        use_ssh_key=False,
        private_key=None,
        private_key_passphrase=None,
        enabled=True,
        owner=None,
    ) -> DatabaseCredential:
        """Create a new Database Credential.

        Database Credentials are used for..ya know..databases

        :param str name: Name of the Database Credential (required)
        :param str database_flavor: Type of database, (default: ``postgres``)
        :param str database: Working database to use, (default: ``postgres``)
        :param str database_host: Database server address, (default: ``localhost``)
        :param int database_port: Port the database server listens on, (default:
            ``5432``)
        :param str database_username: Username to use to connect to the database
        :param str database_password: Password to use to connect to the database
        :param bool use_ssh: Determines if the database will be connected to through a
            tunnel
        :param str ssh_host: The address of the server that the SSH tunnel will connect
            to
        :param str ssh_port: The port the SSH tunnel will use
        :param str ssh_username: Username for the SSH tunnel
        :param str ssh_password: Password for the SSH tunnel
        :param bool use_ssh_key: Allows the credentials to be used
        :param str private_key: SSH private key. Note: No validation will be performed.
        :param str private_key_passphrase: Passphrase for the SSH key
        :param bool enabled: Allows the credentials to be used
        :param Union[User,int,str] owner: Owner of the app. Requires Admin to create for
            other users.

        :returns: DatabaseCredential

        """
        return self._model._create(
            self._credmgr,
            name=name,
            database_flavor=database_flavor,
            database=database,
            database_host=database_host,
            database_port=database_port,
            database_username=database_username,
            database_password=database_password,
            use_ssh=use_ssh,
            ssh_host=ssh_host,
            ssh_port=ssh_port,
            ssh_username=ssh_username,
            ssh_password=ssh_password,
            use_ssh_key=use_ssh_key,
            private_key=private_key,
            private_key_passphrase=private_key_passphrase,
            enabled=enabled,
            owner=owner,
        )


class RefreshTokenHelper(BaseHelper):
    """Interface for interacting with RefreshTokens."""

    _model = RefreshToken

    def __call__(self, id=None, redditor=None, reddit_app_id=None):
        """Fetch an instance of :class:`RefreshToken`."""
        kwargs = {}
        if isinstance(id, str):
            if redditor:
                reddit_app_id = redditor
            redditor = id
            id = None
        if id:
            kwargs["id"] = id
        if redditor:
            kwargs["redditor"] = redditor
        if reddit_app_id:
            kwargs["reddit_app_id"] = reddit_app_id
        return super().__call__(**kwargs)



"""Provide the credmgr class mixins."""

from .base import BaseModel  # NOQA
from .base_app import BaseApp  # NOQA
from .deletable import DeletableMixin  # NOQA
from .editable import EditableMixin  # NOQA
from .owner import OwnerMixin  # NOQA
from .reddit_app import RedditAppMixin  # NOQA
